IMPLEMENTATION MODULE AppointmentBook;


   (*$S-,$T- turn off stack and range checking *) 

   FROM SYSTEM IMPORT ADDRESS, ADR, TSIZE;
   FROM Clock  IMPORT EncodeTime, EncodeDate, DecodeTime, DecodeDate;
   FROM GEMDOS IMPORT ConOut, GetTime, GetDate;
   IMPORT AESApplications;
   IMPORT AESObjects;
   IMPORT AESGraphics;
   IMPORT AESForms;
   IMPORT AESWindows;
   IMPORT Configuration;
   IMPORT Dialog;
   IMPORT File;
   IMPORT GEMAESbase;
   IMPORT Icon;
   IMPORT M2Conversions;
   IMPORT Resource;
   IMPORT Screen;
   IMPORT Text;
   IMPORT Window;
   IMPORT Disk;

   CONST
      Bell    = 07C;
      Max     = 50;
      MaxMsgs = 10;

      FileExistsMsg    =  
"[1][ The AppointmentBook file | already exists. ][Cancel][Replace]"; 

      OpenFailMsgLine1 = "[1][ The AppointmentBook is not | in drive ";
      OpenFailMsgLine2 = 
":. What would you | like to do?][Retry|Create|Cancel]";

      CreateFailMsgLine1 = "[3][ Can't create the file. Make |";
      CreateFailMsgLine2 = " sure that a DeskPak folder | is in drive ";
      CreateFailMsgLine3 = 
": and that the | disk is not write protected. ][Cancel]";

      MissedApptMsg = 
"[3][ You have missed one or more | appointments today ][ OK ]";

      NotInRange = 
"[1][ You have an appointment | which is not in memory ][ Read | Ignore ]";


   TYPE

     ApptTimeType = RECORD
        ApptIndex   : INTEGER;
        TimeIndex   : INTEGER;
     END; (* RECORD *)

     MsgType = ARRAY[0..34] OF CHAR;

     FileRecordType = RECORD
        RecordHeader : File.RecordHeaderType;
        Date   : ARRAY [0..7] OF CHAR;
        Time   : ARRAY [0..7] OF CHAR;
        Msg    : MsgType;
     END; (* RECORD *)

     AppointmentArrayType = ARRAY[1..MaxMsgs] OF MsgType;

     TimeArrayType = ARRAY [1..Max] OF CARDINAL;


   VAR
      AltTimeArray     : TimeArrayType;
      TimeArray        : TimeArrayType;
      AltFirst         : CARDINAL;
      AltLast          : CARDINAL;
      AltNumberOfAppts : CARDINAL;
         (* temporary count of todays # of messages while another *)
         (* date is being processed.                              *)
      AlternateMsgs    : AppointmentArrayType;
         (* temporary array which holds todays messages while another *)
         (* date is being processed.                                  *)
      SelectedItem     : CARDINAL;
         (* index into TodaysMsgs showing which message was picked *)
      ApptTimeArray    : ARRAY[1..MaxMsgs] OF ApptTimeType;
         (* Array of the Resource indexes *)
      BootDate         : CARDINAL;
      DateInMemory     : CARDINAL;
      EntrySize        : LONGCARD;
      FileId           : File.FileIdType;
      FirstMsg         : CARDINAL;
      LastMsg          : CARDINAL;
         (* Range of the Message working set *)
      NotToday         : BOOLEAN;
         (* Flag indicating that another date is being processed *)
      NumberOfAppts    : CARDINAL;
         (* Current number of appointments ( <= Max) *)
      PrevTime         : CARDINAL;
         (* Holds the last time from a MinuteChanged message *)
      RecordId         : File.RecordIdType;
      TodaysMsgs       : AppointmentArrayType;
         (* current array of messages *)
      UpdateRegion     : Screen.Box;
      VirginTool       : BOOLEAN;
      WindowAllocated  : BOOLEAN;
      WindowInfo       : Window.Information;
      WindowRegion     : Screen.Box;
      WindowResource   : INTEGER;
      Drive            : CHAR;
      DriveNumber      : CARDINAL;


   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE GetKey (
      Data    : ADDRESS;
      VAR Key : ARRAY OF CHAR);

   TYPE FileRecordPtr = POINTER TO FileRecordType;

   VAR DataRecord : FileRecordPtr;
     
   BEGIN
     DataRecord := FileRecordPtr (Data);
     Text.Assign (DataRecord^.Date,Key);
   END GetKey;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE GetSubKey (
      Data       : ADDRESS;
      VAR SubKey : ARRAY OF CHAR);

   TYPE FileRecordPtr = POINTER TO FileRecordType;

   VAR DataRecord : FileRecordPtr;
     
   BEGIN
     DataRecord := FileRecordPtr (Data);
     Text.Assign (DataRecord^.Time,SubKey);
   END GetSubKey;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE ConvertToCardinal (
      VAR TheString : ARRAY OF CHAR;
      VAR DoneOk    : BOOLEAN;
      VAR TheNumber : CARDINAL );

   VAR
     CompressedString : Text.String80;
     i      : CARDINAL;
     Length : CARDINAL;
    
   BEGIN
      Length := 0;
      FOR i := 0 TO HIGH(TheString) BY 1 DO
        IF TheString[i] <> ' ' THEN
           CompressedString[Length] := TheString[i];
           INC (Length);
        END; (* IF *)
      END; (* FOR *)
      CompressedString[Length] := CHR (0);
      M2Conversions.ConvertToCardinal (CompressedString, DoneOk, TheNumber);
   END ConvertToCardinal;
 
   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE SelectKey ( Key : INTEGER);

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress(WindowResource, Key);
      ObjectPtr^.State := ObjectPtr^.State + {Icon.Selected};
   END SelectKey;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE DeselectKey ( Key : INTEGER);

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress(WindowResource, Key);
      ObjectPtr^.State := ObjectPtr^.State - {Icon.Selected};
   END DeselectKey;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE EnableKey ( Key : INTEGER);

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress(WindowResource, Key);
      ObjectPtr^.State := ObjectPtr^.State - {Icon.Disabled};
   END EnableKey;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE DisableKey( Key : INTEGER);

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress(WindowResource, Key);
      ObjectPtr^.State := ObjectPtr^.State + {Icon.Disabled};
   END DisableKey;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE DeselectAppts(Index : CARDINAL);

   BEGIN
      IF (Index > 0) AND (Index <= MaxMsgs) THEN
         DeselectKey(ApptTimeArray[Index].ApptIndex);
         DeselectKey(ApptTimeArray[Index].TimeIndex);
      END; (* IF *)
   END DeselectAppts;
  
   (*------------------------- LOCAL ROUTINE --------------------------*)
   (* Put TodaysMsgs in the D11APPT box.                               *)

   PROCEDURE SetUpAppts ();

   TYPE
      TedInfoPtrType = POINTER TO Icon.Tedinfo;
      String8Type    = ARRAY[0..7] OF CHAR;
      String8TypePtr = POINTER TO String8Type;

   VAR
      Start             : CARDINAL;
      Stop              : CARDINAL;
      Index             : CARDINAL;
      ObjectPtr         : Icon.ObjectPtr;
      String8Ptr        : String8TypePtr;
      TedInfoPtr        : TedInfoPtrType;

   BEGIN

     (* Set up the title date *)

     ObjectPtr  := Icon.GetAddress(WindowResource,Resource.D11DATE);
     String8Ptr := String8TypePtr(ObjectPtr^.Spec);
     DecodeDate(DateInMemory, TRUE, String8Ptr^);
 
     IF FirstMsg <= 0 THEN
        Start := 1;
        Stop  := MaxMsgs;
     ELSE
        Start := FirstMsg;
        Stop  := FirstMsg + MaxMsgs - 1;
     END; (* IF *)

     FOR Index := Start TO Stop BY 1 DO
       SetUpAux ( Icon.GetAddress (
          WindowResource,ApptTimeArray[Index - Start + 1].ApptIndex),
          Icon.GetAddress(
          WindowResource,ApptTimeArray[Index - Start + 1].TimeIndex),
          Index);
     END; (* FOR *)
   END SetUpAppts;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE SetUpAux (
      MsgPtr  : Icon.ObjectPtr;
      TimePtr : Icon.ObjectPtr;
      Index   : CARDINAL);

   TYPE
     TedInfoPtrType = POINTER TO Icon.Tedinfo;
     String8Type    = ARRAY[0..8] OF CHAR;
     String8TypePtr = POINTER TO String8Type;
     String35TypePtr= POINTER TO MsgType;

   VAR    
     String8      : String8Type;
     String8Ptr   : String8TypePtr;
     String35Ptr  : String35TypePtr;
     TedinfoPtr   : TedInfoPtrType;

   BEGIN
    TedinfoPtr := TedInfoPtrType (MsgPtr^.Spec);
    String35Ptr:= Text.StringPtr (TedinfoPtr^.Text);

    TedinfoPtr := TedInfoPtrType (TimePtr^.Spec);
    String8Ptr := Text.StringPtr (TedinfoPtr^.Text);

    IF Index <= LastMsg THEN
       Text.Assign(TodaysMsgs[Index - FirstMsg + 1],String35Ptr^);
     
       DecodeTime (TimeArray[Index],TRUE,String8);
       Text.Assign(String8,String8Ptr^);

    ELSE (* Otherwise, the entry is empty *)

       Text.Assign("___________________________________",String35Ptr^);     
       Text.Assign("________",String8Ptr^);
    END; (* IF *)     

   END SetUpAux;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE OpenApptBook ( VAR FileId : File.FileIdType ) : BOOLEAN;

   VAR
      ErrorMsg   : ARRAY[0..150] OF CHAR;
      FileOpened : BOOLEAN;
      Choice     : INTEGER;
      Success    : BOOLEAN;

   BEGIN

      (*--- Try opening the AppointmentBook file ------------------*)

      LOOP

         AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);
         (* See if the medium has changed *)
         Disk.CheckStatus ( DriveNumber );
         File.Open (
             Configuration.ApptFileName, GetKey, GetSubKey, FileId );
         IF File.Status = File.EOK THEN
            FileOpened := TRUE;
            EXIT;
         ELSE
            ErrorMsg := OpenFailMsgLine1;
            Success  := Text.ConcatChar( ErrorMsg, Drive, ErrorMsg);
            Success  := Text.ConcatString ( ErrorMsg, OpenFailMsgLine2, ErrorMsg );
            AESGraphics.GrafMouse (GEMAESbase.Arrow, NIL);
            Choice   := AESForms.FormAlert ( 1, ErrorMsg );
            AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);
            IF Choice = 1 THEN  (* RETRY *)
               (* Stay in loop and try again *)

            ELSIF Choice = 2 THEN  (* CREATE *)

               (* See if the medium has changed *)
               Disk.CheckStatus ( DriveNumber );
               
               (* Make sure the file doesn't exist already *)
               File.Open ( 
                  Configuration.ApptFileName, GetKey, GetSubKey, FileId );
               IF File.Status = File.EOK THEN

                 (* Display a message indicating that the file      *)
                 (* already exists, and do they want to replace it? *)

                 AESGraphics.GrafMouse (GEMAESbase.Arrow, NIL);
                 IF AESForms.FormAlert ( 2,FileExistsMsg) = 1 THEN

                    (* They CANCELed the CREATE; exit successfully *)       

                    FileOpened := TRUE;
                    EXIT;
                 END; (* IF *)
                 AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);

                 (* Close file in preparation for new version *)

                 File.Close(FileId);
               END; (* IF *)

               (* See if the medium has changed *)
               Disk.CheckStatus ( DriveNumber );
               File.Create (
                  Configuration.ApptFileName, 
                  TSIZE (FileRecordType), GetKey, GetSubKey, FileId );
               IF File.Status = File.EOK THEN
                  FileOpened := TRUE;
               ELSE
                  ErrorMsg   := CreateFailMsgLine1;
                  Success    := Text.ConcatString ( 
                      ErrorMsg, CreateFailMsgLine2, ErrorMsg );
                  Success    := Text.ConcatChar ( 
                      ErrorMsg, Drive ,ErrorMsg);
                  Success    := Text.ConcatString ( 
                      ErrorMsg, CreateFailMsgLine3, ErrorMsg );
                  AESGraphics.GrafMouse (GEMAESbase.Arrow, NIL);
                  Choice     := AESForms.FormAlert ( 1, ErrorMsg );
                  FileOpened := FALSE;
               END;
               EXIT;
            ELSE (* Choice = 3 *)
               FileOpened := FALSE;
               EXIT;
            END;
         END;
      END; (* LOOP *)
      AESGraphics.GrafMouse (GEMAESbase.Arrow, NIL);
      RETURN (FileOpened);
   END OpenApptBook;


   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE Hit ( Number,Low,High : CARDINAL) : BOOLEAN;

   BEGIN
     IF (Number >= Low) AND (Number <= High) THEN
        RETURN (TRUE);
     ELSE
        RETURN (FALSE);
     END; (* IF *)
   END Hit;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE AppointmentBookDraw ( Region : Screen.Box );

   BEGIN
      (* Already drawn by the window manager *)
   END AppointmentBookDraw;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE Redraw ( RedrawScreen : BOOLEAN );

   VAR
      VertSliderSize : CARDINAL;
      VertSliderPos  : CARDINAL;

   BEGIN
     IF NumberOfAppts <= MaxMsgs THEN
        VertSliderSize := 1000;
        VertSliderPos  := 0;
     ELSE
        VertSliderSize := CARDINAL (
           LONGCARD (MaxMsgs) *
           LONGCARD (1000) DIV
           LONGCARD (NumberOfAppts));
        IF (FirstMsg > NumberOfAppts - MaxMsgs) THEN
           VertSliderPos  := 1000;
        ELSE
           VertSliderPos  := CARDINAL (
              LONGCARD (FirstMsg-1) *
              LONGCARD (1000) DIV 
              LONGCARD (NumberOfAppts-MaxMsgs));          
        END;
     END;

     IF RedrawScreen THEN
        IF SelectedItem = 0 THEN
           DisableKey(Resource.D11DELET);
           DisableKey(Resource.D11RESCH);
        ELSE
           EnableKey(Resource.D11DELET);
           EnableKey(Resource.D11RESCH);
        END (* IF *);
     END (* IF *);
     AESWindows.WindowSet (
        WindowInfo.Id,
        GEMAESbase.WindowVertSlide,       
        INTEGER(VertSliderPos),0,0,0);
     AESWindows.WindowSet (
        WindowInfo.Id,
        GEMAESbase.VertSliderSize,
        INTEGER(VertSliderSize),0,0,0);

     IF RedrawScreen AND WindowAllocated THEN
        Window.Redraw(WindowInfo.Id,WindowInfo.Borders);
     END; (* IF *)

   END Redraw;

   (*------------------------- LOCAL ROUTINE --------------------------*)
   (* Read in a new date from the file and set it up *)

   PROCEDURE ReadDate (
      NewDate : CARDINAL;
      Switch  : BOOLEAN);

   VAR
      Appt          : FileRecordType;
      CurrentTime   : CARDINAL;
      Date          : CARDINAL;
      MissedAppt    : BOOLEAN;
      FirstRecord   : BOOLEAN;
      Trash         : BOOLEAN;
      FoundKey      : BOOLEAN;
      FoundRecord   : BOOLEAN;
      KeyId         : File.RecordIdType;
      NewAppt       : FileRecordType;
      NewDateString : ARRAY[0..7] OF CHAR;
      Start         : CARDINAL;
      Time          : CARDINAL;
      Today         : CARDINAL;
      Result        : INTEGER;
      
   BEGIN
      AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );
      IF Switch THEN
        (* Save today's messages *)
        AlternateMsgs   := TodaysMsgs;
        AltTimeArray    := TimeArray;
        AltNumberOfAppts:= NumberOfAppts;
        AltFirst        := FirstMsg;
        AltLast         := LastMsg;
        NotToday        := TRUE;
        Start           := 1;
      ELSE
        Start    := FirstMsg;
        IF Start = 0 THEN
           Start := 1;
        END; (* IF *)
      END; (* IF *)
      LastMsg := Start - 1;
       
      IF VirginTool THEN
        GetTime(CurrentTime);
        PrevTime := CurrentTime;
        MissedAppt := FALSE;
      END; (* IF *)     

      NumberOfAppts:= 0;

      IF OpenApptBook(FileId) THEN
       
        AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);
        M2Conversions.ConvertCardinal(NewDate,3,NewDateString);

        (* Go to the specified record, if it exists *)

        File.Find (
           FileId,
           NewDateString,
           "",
           ADR(Appt),
           FoundKey,
           KeyId,
           FoundRecord,
           RecordId );

        IF (File.Status = File.EOK) AND FoundKey THEN
            FirstRecord := TRUE;
            LOOP
               IF FirstRecord THEN
                 File.Seek(FileId,KeyId,ADR(Appt));
                 FirstRecord := FALSE;
               ELSE
                 File.Read(FileId,ADR(Appt),RecordId);
               END; (* IF *)

               IF File.Status # File.EOK THEN
                  EXIT;
               END; (* IF *)
               
               ConvertToCardinal(Appt.Date,Trash,Date);
               ConvertToCardinal(Appt.Time,Trash,Time);
  
               IF (Date # NewDate) OR (NumberOfAppts >= Max) THEN
                  EXIT;
               ELSE

                  (* Otherwise, put the message in the arrays *)

                  INC (NumberOfAppts);
                  TimeArray[NumberOfAppts]  := Time;
                  IF Hit(NumberOfAppts,Start,(Start + MaxMsgs - 1)) THEN
                    TodaysMsgs[NumberOfAppts - Start + 1] := Appt.Msg;
                    INC (LastMsg);
                  END; (* IF *)

                  (* See if they missed the appointment *)

                  IF VirginTool AND (Time < CurrentTime) THEN
                     MissedAppt := TRUE;
                  END; (* IF *)

               END; (* IF *)

             END; (* LOOP *)

             IF VirginTool AND MissedAppt THEN
               AESGraphics.GrafMouse (GEMAESbase.Arrow, NIL);
               Result := AESForms.FormAlert ( 1, MissedApptMsg );
             END; (* IF *)

          END; (* IF *)
 
          File.Close(FileId);

    
       IF NumberOfAppts >= Max THEN
             DisableKey(Resource.D11ADD);
       ELSE
             EnableKey(Resource.D11ADD);
       END; (* IF *)

       DeselectKey(Resource.D11ADD);
       DeselectKey(Resource.D11DELET);
       DeselectKey(Resource.D11RESCH);
       DeselectKey(Resource.D11ERASE);
       DateInMemory := NewDate;

       IF NumberOfAppts = 0 THEN
             FirstMsg := 0;
             LastMsg  := 0;
       ELSE
             FirstMsg := Start;
       END; (* IF *)

       DisableKey(Resource.D11DELET);
       DisableKey(Resource.D11RESCH);

       END; (* IF *)   
       SetUpAppts;

       SelectedItem := 0;
       AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
   END ReadDate;

   (*------------------------- LOCAL ROUTINE --------------------------*)
   (* delete the picked appointment from the file and array *)

   PROCEDURE DeleteAppt (SelectedItem : CARDINAL);

   TYPE MsgBufferType = ARRAY [0..7] OF INTEGER;

   VAR
     i           : CARDINAL;
     MsgBuffer   : MsgBufferType;
     DateString  : ARRAY [0..7] OF CHAR;
     TimeString  : ARRAY [0..7] OF CHAR;
     FoundItem   : BOOLEAN;
     FoundKey    : BOOLEAN;
     FoundRecord : BOOLEAN;
     KeyId       : File.RecordIdType;
     Dummy       : FileRecordType;

   BEGIN
      (* A message must have been previously chosen to delete *)
      IF SelectedItem = 0 THEN
         ConOut( Bell );
      ELSE
         IF OpenApptBook(FileId) THEN  
            AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);
            M2Conversions.ConvertCardinal(DateInMemory,3,DateString);
            M2Conversions.ConvertCardinal(
                TimeArray[SelectedItem + FirstMsg - 1],3,TimeString);

            (* Go to the record *)

            File.Find(
               FileId,
               DateString,
               TimeString,
               ADR(Dummy),
               FoundKey,
               KeyId,
               FoundRecord,
               RecordId);

            FoundItem := FALSE;
            IF FoundKey AND FoundRecord THEN

              (* If there is more than one appt for the same *)
              (* date and time, loop until the correct msg   *)
              (* is found.                                   *)
              IF (Text.Compare(TodaysMsgs[SelectedItem],Dummy.Msg) # Text.Equal) THEN
                File.Seek(FileId,KeyId,ADR(Dummy));

                LOOP
                  IF (File.Status # File.EOK) THEN
                    EXIT;
                  END; (* IF *) 
                  IF (Text.Length (Dummy.Msg) = 0) AND
                     (Text.Length (TodaysMsgs[SelectedItem]) = 0) THEN
                    FoundItem := TRUE;
                    EXIT;
                  ELSIF (Text.Compare(Dummy.Date,DateString) = Text.Equal) AND
                        (Text.Compare(Dummy.Time,TimeString) = Text.Equal) AND
                        (Text.Compare(Dummy.Msg,TodaysMsgs[SelectedItem]) = Text.Equal) THEN
                    FoundItem := TRUE;
                    EXIT;
                  ELSE
                    File.Read(FileId,ADR(Dummy),RecordId);
                  END; (* IF *)
                END; (* LOOP *)

              ELSE
                FoundItem := TRUE;
              END; (* IF *)

              IF FoundItem THEN
                (* delete the node *)
                File.DeleteRecord(FileId,RecordId,ADR(Dummy));
              ELSE
                ConOut(Bell);
              END; (* IF *)
            END; (* IF *)
            File.Close(FileId);

            (* Read the date back into the array *)
            ReadDate(DateInMemory,FALSE);

            (* If the last appointment for a date was deleted, call *)
            (* Calendar so the date won't be highlighted anymore    *)
            IF NumberOfAppts = 0 THEN
               MsgBuffer[0] := Resource.DeletedAppointment;
               MsgBuffer[1] := Resource.ApplicationId;
               MsgBuffer[2] := 0;
               MsgBuffer[3] := 0;
               MsgBuffer[4] := 0;
               MsgBuffer[5] := 0;
               AESApplications.ApplWrite (
                  Resource.ApplicationId,
                  SIZE (MsgBuffer),
                  ADR  (MsgBuffer));
            END; (* IF *)
         END; (* IF *)          
      END; (* IF *)
      DeselectKey(Resource.D11DELET);
   END DeleteAppt;           

   (*------------------------- LOCAL ROUTINE --------------------------*)
   (* Add a new appointment from info gathered from the D12APINF box   *)

   PROCEDURE AddAppt (
      VAR InfoText : ARRAY OF CHAR;
      SetCurrDate  : BOOLEAN) : BOOLEAN;

   TYPE
      TedInfoPtrType = POINTER TO Icon.Tedinfo;
      String5Type    = ARRAY [0..4] OF CHAR;
      String5TypePtr = POINTER TO String5Type;
      String7Type    = ARRAY [0..6] OF CHAR;
      String7TypePtr = POINTER TO String7Type;
      String35Type   = ARRAY [0..34] OF CHAR;
      String35TypePtr= POINTER TO String35Type;
      String8Type    = ARRAY[0..7] OF CHAR;
      String8TypePtr = POINTER TO String8Type;
      MsgBufferType  = ARRAY [0..7] OF INTEGER;

   VAR
     AddRecord   : FileRecordType;
     AddOk       : BOOLEAN;
     AMPtr       : Icon.ObjectPtr;
     PMPtr       : Icon.ObjectPtr;
     CurrentDate : CARDINAL;
     DateString  : String7Type;
     Dummy       : FileRecordType;
     DatePtr     : String8TypePtr;
     FoundKey    : BOOLEAN;
     FoundRecord : BOOLEAN;
     MsgPtr      : String35TypePtr;
     i           : CARDINAL;
     CardDate    : CARDINAL;
     CardTime    : CARDINAL;
     High        : CARDINAL;
     Index       : CARDINAL;
     Today       : CARDINAL;
     KeyId       : File.RecordIdType;
     MsgBuffer   : MsgBufferType;
     ObjectPtr   : Icon.ObjectPtr;
     PM          : BOOLEAN;
     String7Ptr  : String7TypePtr;
     TedinfoPtr  : TedInfoPtrType;
     TimePtr     : String5TypePtr;

   BEGIN
      AddOk  := FALSE;
      (* See if there's room for a new record *)
      AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);
      IF Disk.SpaceAvailable( DriveNumber, EntrySize) THEN
         IF OpenApptBook(FileId) THEN
            AddOk := TRUE;
            AMPtr := Icon.GetAddress ( Resource.D12APINF, Resource.D12AM);
            PMPtr := Icon.GetAddress ( Resource.D12APINF, Resource.D12PM);
            IF SetCurrDate THEN
               GetDate (CurrentDate); 
               DecodeDate ( CurrentDate, FALSE, DateString );
               ObjectPtr := Icon.GetAddress(
                  Resource.D12APINF, Resource.D12DATE);
               TedinfoPtr := TedInfoPtrType (ObjectPtr^.Spec);
               String7Ptr := String7TypePtr (TedinfoPtr^.Text);
               Text.Assign(DateString,String7Ptr^);
            END; (* IF *)

            LOOP (* Loop until valid date and time entered *)

               (* Init the info line *)
               ObjectPtr := Icon.GetAddress (
                  Resource.D12APINF,Resource.D12INFO);
               TedinfoPtr := TedInfoPtrType(ObjectPtr^.Spec);
               MsgPtr     := String35TypePtr(TedinfoPtr^.Text);
               Text.Assign(InfoText,MsgPtr^);
 
               IF Dialog.Display(
                  Resource.D12APINF,
                  Screen.Center,
                  Resource.D12TIME) = Resource.D12OK THEN 

                  (* Output the newly entered information *)
                  (* Time: *)

                  ObjectPtr  := Icon.GetAddress (
                     Resource.D12APINF,
                     Resource.D12TIME);
                  TedinfoPtr := TedInfoPtrType(ObjectPtr^.Spec);
                  TimePtr    := String5TypePtr(TedinfoPtr^.Text);
                  PM         := Icon.Selected IN PMPtr^.State;

                  (* Date: *)
                  ObjectPtr := Icon.GetAddress (
                     Resource.D12APINF,
                     Resource.D12DATE);
                  TedinfoPtr := TedInfoPtrType(ObjectPtr^.Spec);
                  DatePtr    := String8TypePtr(TedinfoPtr^.Text);
       
                  IF (EncodeDate(DatePtr^,CardDate) AND
                     EncodeTime(TimePtr^,PM,CardTime)) THEN
                     EXIT;
                  END; (* IF *)  
               ELSE
                  AddOk := FALSE;
                  EXIT;
               END; (* IF *)
            END; (* LOOP *)

            IF AddOk THEN
               AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);
               (* Appointment: *)
               Text.Assign(MsgPtr^,AddRecord.Msg);

               M2Conversions.ConvertCardinal(CardDate,3,AddRecord.Date);
               M2Conversions.ConvertCardinal(CardTime,3,AddRecord.Time);

               (* See if this is a new appointment for that day *)
               File.Find (
                  FileId,
                  AddRecord.Date,
                  "",
                  ADR(Dummy),
                  FoundKey,
                  KeyId,
                  FoundRecord,
                  RecordId);
               IF (NOT FoundKey) OR (File.Status # File.EOK) OR
                    (NOT FoundRecord) THEN
                  (* Send a message to Calendar *)
                  MsgBuffer[0] := Resource.AddedAppointment;
                  MsgBuffer[1] := Resource.ApplicationId;
                  MsgBuffer[2] := 0;
                  MsgBuffer[3] := 0;
                  MsgBuffer[4] := 0;
                  MsgBuffer[5] := 0;
                  AESApplications.ApplWrite (
                     Resource.ApplicationId,
                     SIZE (MsgBuffer),
                     ADR  (MsgBuffer));
               END; (* IF *)

               (* Insert the record *)
               File.InsertRecord (
                  FileId,
                  AddRecord.Date,
                  AddRecord.Time,
                  ADR(AddRecord),
                  ADR(Dummy),
                  RecordId);
               File.Close(FileId);

               AESGraphics.GrafMouse (GEMAESbase.Arrow, NIL);
               IF File.Status = File.EOK THEN

                  (* If the message is for today, read in the new array *) 

                  IF CardDate = DateInMemory THEN
                     ReadDate(CardDate,FALSE);
                  END; (* IF *)

               END; (* IF *)
            END; (* IF AddOk *)
         END; (* IF OpenFile *)
      END; (* IF Disk.SpaceAvailable *)
      DeselectKey(Resource.D11ADD);
      RETURN (AddOk);
   END AddAppt;

   (*------------------------- LOCAL ROUTINE --------------------------*)
   (* delete the picked message from the current date and set it up at *)
   (* another time.                                                    *)

   PROCEDURE RescheduleAppt ( VAR SelectedItem : CARDINAL );
   
   TYPE
     TedInfoPtrType = POINTER TO Icon.Tedinfo;
     String5Type    = ARRAY[0..4] OF CHAR;
     String5TypePtr = POINTER TO String5Type;
     String6Type    = ARRAY[0..6] OF CHAR;
     String6TypePtr = POINTER TO String6Type;
     String35Type   = ARRAY [0..34] OF CHAR;
     String35TypePtr= POINTER TO String35Type;

   VAR
     DateString  : String6Type;
     Dummy       : FileRecordType;
     FormattedTimeString : ARRAY[0..8] OF CHAR;
     ObjectPtr   : Icon.ObjectPtr;
     SaveRecord  : FileRecordType;
     String5Ptr  : String5TypePtr;
     String6Ptr  : String6TypePtr;
     String35Ptr : String35TypePtr;
     TedinfoPtr  : TedInfoPtrType;
     TextPtr     : Text.StringPtr;
     TimeString  : String5Type;
     Trash       : BOOLEAN;

   BEGIN
      IF SelectedItem = 0 THEN
         ConOut(Bell);
         DeselectKey(Resource.D11RESCH);
      ELSE
         DecodeTime(
            TimeArray[SelectedItem + FirstMsg - 1],
            TRUE,
            FormattedTimeString);
         DecodeTime(
            TimeArray[SelectedItem + FirstMsg - 1],
            FALSE,
            TimeString);
         DecodeDate(DateInMemory,FALSE,DateString);  
 
         (* Save the appointment in case the add fails *)
         M2Conversions.ConvertCardinal(
            TimeArray[SelectedItem + FirstMsg - 1],3,SaveRecord.Time);
         M2Conversions.ConvertCardinal(DateInMemory,3,SaveRecord.Date);
         Text.Assign(TodaysMsgs[SelectedItem],SaveRecord.Msg);
     
         (* set up the information dialog box with the old info *)  

         (* set the buttons *)
         IF FormattedTimeString[6] = "P" THEN
            ObjectPtr := Icon.GetAddress ( 
               Resource.D12APINF, Resource.D12AM);
            ObjectPtr^.State := ObjectPtr^.State - {Icon.Selected};  

            ObjectPtr := Icon.GetAddress ( 
               Resource.D12APINF, Resource.D12PM);
            ObjectPtr^.State := ObjectPtr^.State + {Icon.Selected};
         ELSE
            ObjectPtr := Icon.GetAddress ( 
               Resource.D12APINF, Resource.D12AM);
            ObjectPtr^.State := ObjectPtr^.State + {Icon.Selected};  
 
            ObjectPtr := Icon.GetAddress ( 
               Resource.D12APINF, Resource.D12PM);
            ObjectPtr^.State := ObjectPtr^.State - {Icon.Selected};
         END; (* IF *)
 
         (* set the time *)
         ObjectPtr  := Icon.GetAddress ( 
             Resource.D12APINF, Resource.D12TIME);
         TedinfoPtr := TedInfoPtrType (ObjectPtr^.Spec);
         String5Ptr := String5TypePtr (TedinfoPtr^.Text);
         Text.Assign(TimeString,String5Ptr^); 

         (* set the date *)
         ObjectPtr := Icon.GetAddress ( 
            Resource.D12APINF, Resource.D12DATE);
         TedinfoPtr := TedInfoPtrType (ObjectPtr^.Spec);
         String6Ptr := String6TypePtr (TedinfoPtr^.Text);
         Text.Assign(DateString,String6Ptr^); 

         (* Delete the old message *)
         DeleteAppt(SelectedItem); 

         (* Add the new message *)
         IF NOT AddAppt(SaveRecord.Msg,FALSE) THEN
            IF OpenApptBook(FileId) THEN
               (* call InsertRecord with the deleted record *) 
               (* if AddAppt not successful.                *)
               File.InsertRecord (
                  FileId,
                  SaveRecord.Date,
                  SaveRecord.Time,
                  ADR(SaveRecord),
                  ADR(Dummy),
                  RecordId);
               File.Close(FileId);
               ReadDate(DateInMemory,FALSE);
            END; (* IF *)
         END; (* IF *)
         DeselectKey(Resource.D11RESCH);
         DeselectAppts(SelectedItem);
         SelectedItem := 0;
      END; (* IF *)
   END RescheduleAppt;

   (*-----------------------------------------------------------------*)
   (* Initialization procedure                                        *)

   PROCEDURE Initialize;

   BEGIN
     Drive       := Configuration.ApptFileName[0];
     DriveNumber := ORD (Drive) - ORD ("A") + 1;

     SelectedItem := 0;
     NumberOfAppts:= 0;    
     FirstMsg     := 0;
     LastMsg      := 0;
     NotToday     := FALSE;

     (* Initialize the resource indexes *)
     ApptTimeArray[ 1].ApptIndex := Resource.D11APT1;
     ApptTimeArray[ 1].TimeIndex := Resource.D11TIM1;
     ApptTimeArray[ 2].ApptIndex := Resource.D11APT2;
     ApptTimeArray[ 2].TimeIndex := Resource.D11TIM2;
     ApptTimeArray[ 3].ApptIndex := Resource.D11APT3;
     ApptTimeArray[ 3].TimeIndex := Resource.D11TIM3;
     ApptTimeArray[ 4].ApptIndex := Resource.D11APT4;
     ApptTimeArray[ 4].TimeIndex := Resource.D11TIM4;
     ApptTimeArray[ 5].ApptIndex := Resource.D11APT5;
     ApptTimeArray[ 5].TimeIndex := Resource.D11TIM5;
     ApptTimeArray[ 6].ApptIndex := Resource.D11APT6;
     ApptTimeArray[ 6].TimeIndex := Resource.D11TIM6;
     ApptTimeArray[ 7].ApptIndex := Resource.D11APT7;
     ApptTimeArray[ 7].TimeIndex := Resource.D11TIM7;
     ApptTimeArray[ 8].ApptIndex := Resource.D11APT8;
     ApptTimeArray[ 8].TimeIndex := Resource.D11TIM8;
     ApptTimeArray[ 9].ApptIndex := Resource.D11APT9;
     ApptTimeArray[ 9].TimeIndex := Resource.D11TIM9;
     ApptTimeArray[10].ApptIndex := Resource.D11APT10;
     ApptTimeArray[10].TimeIndex := Resource.D11TIM10;

     GetDate(BootDate);
     DateInMemory := BootDate;
     ReadDate(DateInMemory,FALSE);
     VirginTool := FALSE;
   END Initialize;

   (*-----------------------------------------------------------------*)

   PROCEDURE EraseOld(Date : CARDINAL);

   CONST
   EraseCheckMsg = 
"[2][ Erase all appointments prior | to the current date/time ][OK|CANCEL]";
   TYPE MsgBufferType = ARRAY [0..7] OF INTEGER;   
 
   VAR
      CardTime,
      CardDate    : CARDINAL;
      CurrentTime : CARDINAL;
      MsgBuffer   : MsgBufferType;
      PrevDate    : CARDINAL;
      Trash       : BOOLEAN;
      TempRecord  : FileRecordType;
      Dummy       : FileRecordType;

   BEGIN
     IF AESForms.FormAlert(2, EraseCheckMsg) = 1 THEN
      IF OpenApptBook(FileId) THEN
 
         AESGraphics.GrafMouse (GEMAESbase.HourGlass, NIL);
         GetTime(CurrentTime);

         (* Go to the beginning of the file *)

         RecordId := FileId.FirstRecordId;
         File.Read(FileId,ADR(TempRecord),RecordId);
         ConvertToCardinal(TempRecord.Date,Trash,CardDate);
         ConvertToCardinal(TempRecord.Time,Trash,CardTime);
         PrevDate := CardDate;

         (* Erase all appointments up to the current date/time *)
         LOOP
           IF File.Status # File.EOK THEN
             EXIT;
           ELSIF CardDate > Date THEN
             EXIT;
           ELSIF (CardDate = Date) AND (CardTime > CurrentTime) THEN
             EXIT;
           END; (* IF *)            

           File.DeleteRecord(FileId,RecordId,ADR(TempRecord));
           File.Read(FileId,ADR(TempRecord),RecordId); 
           ConvertToCardinal(TempRecord.Time,Trash,CardTime);
           ConvertToCardinal(TempRecord.Date,Trash,CardDate);
         END; (* LOOP *)

         (* Inform the calendar that appointments have been deleted *)

         MsgBuffer[0] := Resource.DeletedAppointment;
         MsgBuffer[1] := Resource.ApplicationId;
         MsgBuffer[2] := 0;
         MsgBuffer[3] := 0;
         MsgBuffer[4] := 0;
         MsgBuffer[5] := 0;
         AESApplications.ApplWrite (
            Resource.ApplicationId,
            SIZE (MsgBuffer),
            ADR  (MsgBuffer));
           
         File.Close(FileId);
         AESGraphics.GrafMouse (GEMAESbase.Arrow, NIL);
         ReadDate(Date,FALSE);
      END; (* IF *)
    END; (* IF *)
    DeselectKey(Resource.D11ERASE);
   END EraseOld;

   (*------------------------------------------------------------------*)

   PROCEDURE OwnsWindow ( WindowId : INTEGER ) : BOOLEAN;

   BEGIN
     IF WindowAllocated THEN
         RETURN (WindowId = WindowInfo.Id);
      ELSE
         RETURN (FALSE);
      END;
   END OwnsWindow;

   (*------------------------------------------------------------------*)

   PROCEDURE Open ( UseCurrentDate : BOOLEAN;
                    NewDate        : CARDINAL);

   TYPE
     TedInfoPtrType = POINTER TO Icon.Tedinfo;
     String7Type    = ARRAY [0..6] OF CHAR;
     String7TypePtr = POINTER TO String7Type;

   VAR
      AMPtr             : Icon.ObjectPtr;
      PMPtr             : Icon.ObjectPtr;      
      VirtualRegionSize : Window.Area;
      ExitObject        : INTEGER;
      RedrawScreen      : BOOLEAN;
      String7Ptr        : String7TypePtr;
      DateString        : String7Type;
      TedinfoPtr        : TedInfoPtrType;
      ObjectPtr         : Icon.ObjectPtr;

   BEGIN
      SelectedItem := 0;
      IF WindowAllocated THEN
         RedrawScreen := TRUE;
         Window.Top ( WindowInfo.Id );
      ELSE
         RedrawScreen := FALSE;
         Icon.GetRegion ( WindowResource, 0, WindowRegion );
         WindowRegion.Origin.X := 
            Screen.Center.Origin.X - (WindowRegion.Size.Width DIV 2);
         WindowRegion.Origin.Y := 
            Screen.Center.Origin.Y - (WindowRegion.Size.Height DIV 2);
         UpdateRegion := WindowRegion;
         VirtualRegionSize.Width  :=
            Window.Pixel (WindowRegion.Size.Width);
         VirtualRegionSize.Height :=
            Window.Pixel (WindowRegion.Size.Height);
         IF Window.Open (
               Screen.Center,
               Window.InformationPtr (ADR (WindowInfo)),
               " Appointment Book ",
               "",          (* No information line *)
               WindowRegion,
               VirtualRegionSize,
               GEMAESbase.Name + 
               GEMAESbase.Closer + 
               GEMAESbase.Mover +
               GEMAESbase.UpArrow +
               GEMAESbase.DownArrow +
               GEMAESbase.VertSlide,
               0, 1,    (* White Background *)
               WindowResource,
               AppointmentBookDraw ) THEN
            WindowAllocated := TRUE;
         ELSE
            Window.Unavailable;
            WindowAllocated := FALSE;
         END; (* IF *)
      END; (* IF *)

      IF WindowAllocated THEN
         IF UseCurrentDate THEN

            (* if another date is currently in the array, switch it *)

            IF NotToday THEN
               TodaysMsgs    := AlternateMsgs;
               TimeArray     := AltTimeArray;
               NumberOfAppts := AltNumberOfAppts;
               FirstMsg      := AltFirst;
               LastMsg       := AltLast;
               DateInMemory  := BootDate;
               NotToday      := FALSE;
            END; (* IF *)
            ReadDate ( DateInMemory, FALSE );
            DecodeDate ( DateInMemory, FALSE, DateString );

         ELSE 
            ReadDate ( NewDate, TRUE );
               DecodeDate (NewDate, FALSE, DateString );

         END; (* IF *)          

         (* Set Default to AM *)
         AMPtr := Icon.GetAddress ( Resource.D12APINF, Resource.D12AM);
         PMPtr := Icon.GetAddress ( Resource.D12APINF, Resource.D12PM);
         AMPtr^.State := AMPtr^.State + {Icon.Selected};
         PMPtr^.State := PMPtr^.State - {Icon.Selected};

         (* Set the default date *)
         ObjectPtr := Icon.GetAddress(
             Resource.D12APINF, Resource.D12DATE);
         TedinfoPtr := TedInfoPtrType (ObjectPtr^.Spec);
         String7Ptr := String7TypePtr (TedinfoPtr^.Text);
         Text.Assign(DateString,String7Ptr^);

         Redraw ( RedrawScreen );

      END; (* IF *)
   END Open;

   (*------------------------------------------------------------------*)

   PROCEDURE Close;
 
   BEGIN
     IF WindowAllocated THEN
       DeselectAppts(SelectedItem);
       Window.Close (WindowInfo.Id, Screen.Center);
       WindowAllocated := FALSE;
     END; (* IF *)
   END Close;

   (*------------------------------------------------------------------*)

   PROCEDURE ProcessMessageEvent ( VAR MsgBuffer : ARRAY OF INTEGER ); 

   TYPE
      TedInfoPtrType  = POINTER TO Icon.Tedinfo;
      String35Type    = ARRAY[0..34] OF CHAR;
      String35TypePtr = POINTER TO String35Type;
     
   VAR
      AlarmMsg    : MsgType;
      TimeString  : ARRAY [0..7] OF CHAR;
      DateString  : ARRAY [0..7] OF CHAR;
      DateInMemory: CARDINAL;
      FoundOk     : BOOLEAN;
      High        : CARDINAL;
      IntBuffer   : ARRAY[0..18] OF INTEGER;
      MatchTime   : CARDINAL;
      Message     : Text.String80;
      MsgPtr      : String35TypePtr;
      NameAddress : ADDRESS;
      TextBuffer  : ARRAY [0..34] OF CHAR;
      NameBuffer  : ARRAY [0..34] OF CHAR;
      ObjectPtr   : Icon.ObjectPtr;
      OldDate     : CARDINAL;
      Position    : CARDINAL;
      Result      : INTEGER;
      TempIndex   : INTEGER;
      TedInfoPtr  : TedInfoPtrType;
      TempRecord  : FileRecordType;
      Time        : CARDINAL;
      i           : CARDINAL;
      Trash       : BOOLEAN;
      FoundKey    : BOOLEAN;
      FoundRecord : BOOLEAN;
      KeyId       : File.RecordIdType;

   BEGIN
      IF MsgBuffer[0] = GEMAESbase.AccessoryClose THEN
        IF WindowAllocated THEN
           Window.Close (0, Screen.Center );
           WindowAllocated := FALSE;
        END;

      ELSIF MsgBuffer[0] = GEMAESbase.WindowArrowed THEN

         IF MsgBuffer[4] = 0 THEN (* Page Up *)

            (* Retreat by up to MaxMsgs positions *)

            IF (FirstMsg > 1) AND (NumberOfAppts > MaxMsgs) THEN
               DeselectAppts(SelectedItem);
               SelectedItem := 0;
               IF MaxMsgs >= FirstMsg THEN
                 FirstMsg := 1;
               ELSE
                 FirstMsg := FirstMsg - MaxMsgs + 1;
               END; (* IF *)
               ReadDate(DateInMemory, FALSE);
               Redraw ( TRUE );
            ELSE
               ConOut( Bell );
            END; (* IF *)

         ELSIF MsgBuffer[4] = 1 THEN (* Page Down *)

            (* Advance by up to MaxMsgs positions *)
 
            IF (FirstMsg < (NumberOfAppts - MaxMsgs + 1)) AND
                (NumberOfAppts > MaxMsgs) THEN
               DeselectAppts(SelectedItem);
               SelectedItem := 0;
               FirstMsg := FirstMsg + MaxMsgs;
               IF FirstMsg > (NumberOfAppts - MaxMsgs + 1) THEN
                  FirstMsg := NumberOfAppts - MaxMsgs + 1;
               END; (* IF *)
               ReadDate(DateInMemory,FALSE);
               Redraw ( TRUE );
            ELSE
               ConOut( Bell );
            END; (* IF *)

         ELSIF MsgBuffer[4] = 2 THEN (* Row Up *) 
            IF FirstMsg > 1 THEN
               DeselectAppts(SelectedItem);
               SelectedItem := 0;
               DEC (FirstMsg);
               ReadDate(DateInMemory,FALSE);
               Redraw ( TRUE );
            ELSE
               ConOut( Bell );
            END; (* IF *)

         ELSIF MsgBuffer[4] = 3 THEN (* Row Down *)
            IF (FirstMsg < (NumberOfAppts - MaxMsgs + 1)) AND 
                (NumberOfAppts > MaxMsgs) THEN
              DeselectAppts(SelectedItem);
              SelectedItem := 0;
              INC (FirstMsg);
              ReadDate(DateInMemory,FALSE);
              Redraw ( TRUE );
            ELSE
              ConOut(Bell);
            END; (* IF *)
            
         END; (* IF *)
         
      ELSIF MsgBuffer[0] = GEMAESbase.WindowVertSlided THEN
      
         (* Process a vertical slide...the position is in 0..1000 *)
       IF (NumberOfAppts > MaxMsgs) THEN
         DeselectAppts(SelectedItem);
         SelectedItem := 0;
         (* Get index into the Time array *)
         FirstMsg := CARDINAL(
            LONGCARD(NumberOfAppts - MaxMsgs) * 
            LONGCARD(MsgBuffer[4]) DIV
            LONGCARD(1000)) + 1;

         (* Read the next 10 records into memory *)
         ReadDate(DateInMemory,FALSE);
         Redraw ( TRUE );
       END; (* IF *)

      ELSIF MsgBuffer[0] = Resource.MinuteChanged THEN
         GetTime(Time);
         IF NotToday THEN
            High := AltNumberOfAppts;
         ELSE
            High := NumberOfAppts;
         END; (* IF *)
         FOR i := 1 TO High BY 1 DO
            IF NotToday THEN
               MatchTime := AltTimeArray[i];
            ELSE
               MatchTime := TimeArray[i];
            END; (* IF *)

            (* If a time matches, create a dialog box & display it.  *)
            (* Check the correct array, depending on whether another *)
            (* date is currently being displayed.                    *)
          
            FoundOk := FALSE;
            IF (MatchTime > PrevTime) AND (MatchTime <= Time) THEN
               M2Conversions.ConvertCardinal(MatchTime,3,TimeString);

               (* See if the appointment is in memory: *)

               IF NOT Hit(i,FirstMsg,FirstMsg + MaxMsgs - 1) THEN

                  (* If they want to read the message, read it in *)

                  IF AESForms.FormAlert(1, NotInRange) = 1 THEN
                     IF OpenApptBook(FileId) THEN
                        IF NotToday THEN
                           M2Conversions.ConvertCardinal ( 
                               BootDate,3,DateString); 
                        ELSE
                           M2Conversions.ConvertCardinal ( 
                               DateInMemory,3,DateString);
                        END; (* IF *) 
                        File.Find ( 
                           FileId,
                           DateString,
                           TimeString,
                           ADR(TempRecord),
                           FoundKey,
                           KeyId,
                           FoundRecord,
                           RecordId );
                        File.Seek(FileId,KeyId,ADR(TempRecord));                  
                        File.Close(FileId);
                        FoundOk := TRUE;
                        Text.Assign(TempRecord.Msg,AlarmMsg);
                     END; (* IF *)
                  END; (* IF *)

               ELSIF NotToday THEN
                  AlarmMsg := AlternateMsgs[i];
                  FoundOk := TRUE;
               ELSE
                  AlarmMsg := TodaysMsgs[i];
                  FoundOk := TRUE;
               END; (* IF *)                       

               IF FoundOk THEN
                  Trash := Text.ConcatString("[1][",AlarmMsg,Message);
                  Trash := Text.ConcatString(Message,"][ OK ]",Message);                                      
                  ConOut (Bell);
                  Result := AESForms.FormAlert ( 1, Message );
               END; (* IF *)     
            END; (* IF *)
         END; (* FOR *)               
         PrevTime := Time;

      ELSIF MsgBuffer[0] = Resource.DateChanged THEN

         (* if the date has changed and the current date is in *)
         (* memory, reset the array.                           *)
 
         GetDate(BootDate);
         IF (NOT NotToday) THEN
           DateInMemory := BootDate;
           IF WindowAllocated THEN
               DeselectAppts(SelectedItem);
               SelectedItem := 0;
           END; (* IF *)
           ReadDate(DateInMemory,FALSE);
           IF WindowAllocated THEN
             Redraw ( TRUE );
           END; (* IF *)
         ELSE
           (* Set up the new date in the Alternate array: *)
           (* First, save whatever is currently being viewed *)
           (* and put the new date in TodaysMsgs *)
           OldDate := DateInMemory;
           ReadDate(BootDate,TRUE);
           (* Now read the date being viewed back into TodaysMsgs and *)
           (* put the new date in the alternate array *)
           ReadDate(OldDate,TRUE);
         END; (* IF *)

      ELSIF MsgBuffer[0] = Resource.CallBack THEN
         (* Get name from PhoneBook *)
         Text.Assign ( "Call ", TextBuffer );
         Trash := TRUE;
         NameAddress :=
            ADDRESS (CARDINAL (MsgBuffer[3])) * 10000H +
            ADDRESS (CARDINAL (MsgBuffer[4]));
         IF NameAddress # 0 THEN
            MsgPtr := String35TypePtr (NameAddress);
            Text.Assign ( MsgPtr^, NameBuffer );
            IF Text.Length ( NameBuffer ) > 30 THEN
               Trash := Text.SliceString ( NameBuffer, 0, 29, NameBuffer );
            END;
            Trash := Text.ConcatString ( TextBuffer, NameBuffer, TextBuffer );
         END; (* IF *)
         IF Trash THEN
            (* Set up D12APINFO here *)
            Trash := AddAppt ( TextBuffer, TRUE );
            ReadDate ( DateInMemory, FALSE );
            Redraw ( TRUE );          
         END; (* IF *)
      END; (* IF *)
   END ProcessMessageEvent;

   (*------------------------------------------------------------------*)

   PROCEDURE ProcessButtonEvent (
      MouseButton : INTEGER;
      ModifierKey : INTEGER;
      Clicks      : INTEGER );

   VAR
      MouseLocation : Screen.PixelCoordinate;
      MouseState    : INTEGER;
      KeyboardState : INTEGER;
      Object        : INTEGER;
      ObjectPtr     : Icon.ObjectPtr;
      i             : CARDINAL;
      Trash         : BOOLEAN;
      SaveItem      : CARDINAL;

   BEGIN
      AESGraphics.GrafMouseKeyboardState (
         MouseLocation.X,
         MouseLocation.Y,
         MouseState,
         KeyboardState );
      IF Icon.Find ( WindowResource, MouseLocation, Object ) THEN
         ObjectPtr := Icon.GetAddress(WindowResource,Object);
         IF Icon.Disabled IN ObjectPtr^.State THEN
            ConOut( Bell );
         ELSE
            IF ObjectPtr^.Type = GEMAESbase.GraphicButton THEN
               ObjectPtr^.State := ObjectPtr^.State + {Icon.Selected};
               Icon.Display(WindowResource, Object, WindowInfo.Borders);
               ObjectPtr^.State := ObjectPtr^.State - {Icon.Selected};
            END; (* IF *)
            DeselectAppts(SelectedItem);

            IF Object = Resource.D11DELET THEN
               DeleteAppt (SelectedItem);
               SelectedItem := 0;
               Redraw ( TRUE );
            ELSIF Object = Resource.D11ADD THEN
               SaveItem := SelectedItem;
               SelectedItem := 0;
               IF AddAppt ("",FALSE) THEN
                 Redraw ( TRUE );
               ELSIF SaveItem # 0 THEN
                 DisableKey(Resource.D11DELET);
                 DisableKey(Resource.D11RESCH);
                 Icon.Display (
                     WindowResource,
                     Resource.D11ADD,
                     WindowInfo.Borders);
                 Icon.Display (
                     WindowResource,
                     Resource.D11DELET,
                     WindowInfo.Borders);
                 Icon.Display (
                     WindowResource,
                     Resource.D11RESCH,
                     WindowInfo.Borders);
                 Icon.Display (
                     WindowResource,
                     ApptTimeArray[SaveItem].TimeIndex,
                     WindowInfo.Borders);
                 Icon.Display (
                     WindowResource,
                     ApptTimeArray[SaveItem].ApptIndex,
                     WindowInfo.Borders);
               END; (* IF *)
            ELSIF Object = Resource.D11RESCH THEN
               RescheduleAppt (SelectedItem);
               SelectedItem := 0;
               Redraw ( TRUE );
            ELSIF Object = Resource.D11ERASE THEN
               EraseOld(DateInMemory);
               SelectedItem := 0;
               Redraw ( TRUE );
            ELSE (* Check the ApptTimeArray *)
               IF Hit(SelectedItem,1,MaxMsgs) THEN
                 (* Appt has already been deselected; display it *)
                 Icon.Display (
                   WindowResource,
                   ApptTimeArray[SelectedItem].TimeIndex,
                   WindowInfo.Borders);
                 Icon.Display (
                   WindowResource,
                   ApptTimeArray[SelectedItem].ApptIndex,
                   WindowInfo.Borders);
               END; (* IF *)

               SelectedItem := 0;
               i := 1;      
               LOOP
                  IF (((Object = ApptTimeArray[i].TimeIndex) OR
                        (Object = ApptTimeArray[i].ApptIndex)) AND
                        (i <= NumberOfAppts)) THEN
                     SelectedItem := i;
                     SelectKey(ApptTimeArray[i].ApptIndex);
                     SelectKey(ApptTimeArray[i].TimeIndex);
                     Icon.Display (
                        WindowResource,
                        ApptTimeArray[i].TimeIndex,
                        WindowInfo.Borders);
                     Icon.Display (
                        WindowResource,
                        ApptTimeArray[i].ApptIndex,
                        WindowInfo.Borders);
                     EXIT;
                  END; (* IF *)
                  INC (i);
                  IF (i > MaxMsgs) OR (i > NumberOfAppts) THEN
                     ConOut( Bell );
                     EXIT;
                  END; (* IF *)
               END; (* LOOP *)

               IF SelectedItem = 0 THEN
                  DisableKey(Resource.D11DELET);
                  DisableKey(Resource.D11RESCH);
               ELSE
                  EnableKey(Resource.D11DELET);
                  EnableKey(Resource.D11RESCH);
               END; (* IF *)
               Icon.Display (
                  WindowResource,
                  Resource.D11DELET,
                  WindowInfo.Borders );
               Icon.Display (
                  WindowResource,
                  Resource.D11RESCH,
                  WindowInfo.Borders);
            END; (* IF *)
         END; (* IF *)
      END;         
   END ProcessButtonEvent;

   (*------------------------------------------------------------------*)

   PROCEDURE ProcessKeyboardEvent (
      Key         : INTEGER;
      ModifierKey : INTEGER );

   BEGIN
     ConOut( Bell );
   END ProcessKeyboardEvent;

   (*------------------------------------------------------------------*)

   PROCEDURE DatesWithAppts(
      Month         : CARDINAL;
      Year          : CARDINAL;
      VAR DateArray : CalenderType);

   VAR
     CardDate         : CARDINAL;
     Day              : CARDINAL;
     NextMonth        : CARDINAL;
     FirstOfThisMonth : CARDINAL;
     FirstOfNextMonth : CARDINAL;
     FoundKey         : BOOLEAN;
     FoundRecord      : BOOLEAN;
     KeyId            : File.RecordIdType;
     i                : INTEGER;
     TempRecord       : FileRecordType;
     Trash            : BOOLEAN;

   BEGIN
      FOR i := 1 TO 31 BY 1 DO
         DateArray[i] := FALSE;
      END; (* FOR *)

      IF OpenApptBook(FileId) THEN        
         (* create the first of this and next month *)
         FirstOfThisMonth := ((Year - 1980) * 512) + (Month * 32) + 1;
     
         IF Month = 12 THEN
            NextMonth := 1;
            INC (Year);
         ELSE
            NextMonth := Month + 1;
         END; (* IF *)
         FirstOfNextMonth := ((Year - 1980) * 512) + (NextMonth * 32) + 1;

         (* go to the first record *)
         File.Read ( FileId, ADR(TempRecord), RecordId);

         ConvertToCardinal(TempRecord.Date,Trash,CardDate);

         (* Loop until a date greater than the first of next *)
         (* month is read.                                   *)

         WHILE ((File.Status = File.EOK) AND 
               (CardDate < FirstOfNextMonth)) DO
            IF CardDate >= FirstOfThisMonth THEN
               (* extract the DAY from the date *)
               Day := CardDate MOD 32;
               DateArray[Day] := TRUE;          
            END; (* IF *)
            File.Read(FileId,ADR(TempRecord),RecordId);
            IF File.Status = File.EOK THEN
               ConvertToCardinal(TempRecord.Date,Trash,CardDate);
            END; (* IF *)
         END; (* WHILE *)
         File.Close(FileId);
      END; (* IF *)
   END DatesWithAppts;


BEGIN
  EntrySize       := LONGCARD (TSIZE (FileRecordType));
  VirginTool      := TRUE;
  WindowAllocated := FALSE;
  WindowResource  := Resource.D11APPT;
END AppointmentBook.

