IMPLEMENTATION MODULE Calculator;


   (*$S-,$T- turn off stack and range checking *) 

   FROM SYSTEM IMPORT ADR;
   IMPORT GEMDOS;
   IMPORT GEMAESbase;
   IMPORT AESGraphics;
   IMPORT Text;
   IMPORT Resource;
   IMPORT Screen;
   IMPORT Dialog;
   IMPORT Icon;
   IMPORT Window;
   IMPORT Program;
   IMPORT Keypad;
   IMPORT Expression;


   CONST Bell = 07C;

   VAR
      WindowAllocated : BOOLEAN;
      WindowInfo      : Window.Information;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE CalculatorDraw ( Region : Screen.Box );

   BEGIN
      ;   (* Already drawn by the window manager *)
   END CalculatorDraw;

   (*------------------------- LOCAL ROUTINE --------------------------*)

   PROCEDURE ProcessKey ( Key : INTEGER );

   CONST ShowKeys = TRUE;

   VAR
      ObjectPtr        : Icon.ObjectPtr;
      Numberbase       : INTEGER;
      NumberWasEntered : BOOLEAN;
      Number           : Text.String80;

   BEGIN

      (*--- Ignore the key if it is disabled --------------------------*)

      ObjectPtr := Icon.GetAddress ( Resource.D9CALC, Key );
      IF (ObjectPtr^.Type = GEMAESbase.GraphicButton) AND
            NOT (Icon.Disabled IN ObjectPtr^.State) THEN

         (*--- The key is enabled, so process it ----------------------*)

         IF NOT Program.ProcessKey ( Key, WindowInfo.Borders ) THEN

            (*--- If the program didnt gobble the key, process it -----*)
            (*--- interactively.                                  -----*)

            IF Keypad.ProcessKey (
                  Key,
                  WindowInfo.Borders,
                  ShowKeys,
                  Number ) THEN
               Expression.ProcessKey ( Key, WindowInfo.Borders, Number );
            END;
         END;
      ELSE
         GEMDOS.ConOut ( Bell );
      END;
   END ProcessKey;

   (*------------------------------------------------------------------*)

   PROCEDURE Initialize ();

   BEGIN
      Keypad.Initialize;
      Program.Initialize;
   END Initialize;

   (*------------------------------------------------------------------*)

   PROCEDURE OwnsWindow ( WindowId : INTEGER ) : BOOLEAN;

   BEGIN
      IF WindowAllocated THEN
         RETURN (WindowId = WindowInfo.Id);
      ELSE
         RETURN (FALSE);
      END;
   END OwnsWindow;

   (*------------------------------------------------------------------*)

   PROCEDURE Open;

   VAR
      WindowRegion      : Screen.Box;
      VirtualRegionSize : Window.Area;

   BEGIN
      IF WindowAllocated THEN
         Window.Top ( WindowInfo.Id );
      ELSE
         Icon.GetRegion ( Resource.D9CALC, 0, WindowRegion );
         WindowRegion.Origin.X := 
            Screen.Center.Origin.X - (WindowRegion.Size.Width DIV 2);
         WindowRegion.Origin.Y := 
            Screen.Center.Origin.Y - (WindowRegion.Size.Height DIV 2);
         VirtualRegionSize.Width  :=
            Window.Pixel (WindowRegion.Size.Width);
         VirtualRegionSize.Height :=
            Window.Pixel (WindowRegion.Size.Height);
         IF Window.Open (
               Screen.Center,
               Window.InformationPtr (ADR (WindowInfo)),
               " Calculator ",
               "",               (* No information line *)
               WindowRegion,
               VirtualRegionSize,
               GEMAESbase.Name + GEMAESbase.Closer + GEMAESbase.Mover,
               0, 1,             (* White Background *)
               Resource.D9CALC,
               CalculatorDraw ) THEN
            WindowAllocated := TRUE;
         ELSE
            Window.Unavailable;
            WindowAllocated := FALSE;
         END;
      END;
   END Open;

   (*------------------------------------------------------------------*)

   PROCEDURE Close;

   BEGIN
      IF WindowAllocated THEN
         Window.Close ( WindowInfo.Id, Screen.Center );
         WindowAllocated := FALSE;
      END;
   END Close;

   (*------------------------------------------------------------------*)

   PROCEDURE ProcessMessageEvent ( VAR MsgBuffer : ARRAY OF INTEGER ); 

   BEGIN
      IF MsgBuffer[0] = GEMAESbase.AccessoryClose THEN
         IF WindowAllocated THEN
            Window.Close ( 0, Screen.Center );
            WindowAllocated := FALSE;
         END;
      ELSE
         GEMDOS.ConOut ( Bell );
      END;
   END ProcessMessageEvent;

   (*------------------------------------------------------------------*)

   PROCEDURE ProcessButtonEvent ( MouseButton : INTEGER;
                                  ModifierKey : INTEGER;
                                  Clicks      : INTEGER );

   VAR
      MouseLocation    : Screen.PixelCoordinate;
      MouseState       : INTEGER;
      KeyboardState    : INTEGER;
      Key              : INTEGER;

   BEGIN
      AESGraphics.GrafMouseKeyboardState (
         MouseLocation.X, MouseLocation.Y, MouseState, KeyboardState );
      IF Icon.Find ( Resource.D9CALC, MouseLocation, Key ) THEN
         ProcessKey ( Key );
      ELSE
         GEMDOS.ConOut ( Bell );
      END;
   END ProcessButtonEvent;

   (*------------------------------------------------------------------*)

   PROCEDURE ProcessKeyboardEvent ( Key         : INTEGER;
                                    ModifierKey : INTEGER );

   CONST
      Undefined = 0;
      CR        = 15C;
      Clr       = 47H;
      F1        = 3BH;
      F2        = 3CH;
      F3        = 3DH;
      F4        = 3EH;
      F5        = 3FH;
      F6        = 40H;
      F7        = 41H;
      F8        = 42H;
      F9        = 43H;
      F10       = 44H;

   VAR
      Ch        : CHAR;
      KeyObject : INTEGER;
      HighByte  : INTEGER;

   BEGIN
      HighByte := Key DIV 256;
      Ch       := CHAR (Key MOD 256);
      IF Ch = '0' THEN
         KeyObject := Resource.D9KEY0;
      ELSIF Ch = '1' THEN
         KeyObject := Resource.D9KEY1;
      ELSIF Ch = '2' THEN
         KeyObject := Resource.D9KEY2;
      ELSIF Ch = '3' THEN
         KeyObject := Resource.D9KEY3;
      ELSIF Ch = '4' THEN
         KeyObject := Resource.D9KEY4;
      ELSIF Ch = '5' THEN
         KeyObject := Resource.D9KEY5;
      ELSIF Ch = '6' THEN
         KeyObject := Resource.D9KEY6;
      ELSIF Ch = '7' THEN
         KeyObject := Resource.D9KEY7;
      ELSIF Ch = '8' THEN
         KeyObject := Resource.D9KEY8;
      ELSIF Ch = '9' THEN
         KeyObject := Resource.D9KEY9;
      ELSIF (Ch = 'A') OR (Ch = 'a') THEN
         KeyObject := Resource.D9KEYA;
      ELSIF (Ch = 'B') OR (Ch = 'b') THEN
         KeyObject := Resource.D9KEYB;
      ELSIF (Ch = 'C') OR (Ch = 'c') THEN
         KeyObject := Resource.D9KEYC;
      ELSIF (Ch = 'D') OR (Ch = 'd') THEN
         KeyObject := Resource.D9KEYD;
      ELSIF (Ch = 'E') OR (Ch = 'e') THEN
         KeyObject := Resource.D9KEYE;
      ELSIF (Ch = 'F') OR (Ch = 'f') THEN
         KeyObject := Resource.D9KEYF;
      ELSIF Ch = '.' THEN
         KeyObject := Resource.D9POINT;
      ELSIF Ch = '+' THEN
         KeyObject := Resource.D9PLUS;
      ELSIF Ch = '-' THEN
         KeyObject := Resource.D9MINUS;
      ELSIF Ch = '*' THEN
         KeyObject := Resource.D9MULT;
      ELSIF Ch = '/' THEN
         KeyObject := Resource.D9DIV;
      ELSIF Ch = '(' THEN
         KeyObject := Resource.D9LPAREN;
      ELSIF Ch = ')' THEN
         KeyObject := Resource.D9RPAREN;
      ELSIF Ch = CR THEN
         KeyObject := Resource.D9EQUALS;
      ELSIF HighByte = Clr THEN
         KeyObject := Resource.D9CE;
      ELSIF HighByte = F1 THEN
         KeyObject := Resource.D9F1;  
      ELSIF HighByte = F2 THEN
         KeyObject := Resource.D9F2;  
      ELSIF HighByte = F3 THEN
         KeyObject := Resource.D9F3;  
      ELSIF HighByte = F4 THEN
         KeyObject := Resource.D9F4;  
      ELSIF HighByte = F5 THEN
         KeyObject := Resource.D9F5;  
      ELSIF HighByte = F6 THEN
         KeyObject := Resource.D9F6;  
      ELSIF HighByte = F7 THEN
         KeyObject := Resource.D9F7;  
      ELSIF HighByte = F8 THEN
         KeyObject := Resource.D9F8;  
      ELSIF HighByte = F9 THEN
         KeyObject := Resource.D9F9;  
      ELSIF HighByte = F10 THEN
         KeyObject := Resource.D9F10;  
      ELSE 
         KeyObject := Undefined;
         GEMDOS.ConOut ( Bell );
      END;
      IF KeyObject <> Undefined THEN
         ProcessKey ( KeyObject );
      END;
   END ProcessKeyboardEvent;

BEGIN
   WindowAllocated := FALSE;
END Calculator.

