IMPLEMENTATION MODULE Calendar;


   (*$S-,$T- turn off stack and range checking *) 

   FROM SYSTEM IMPORT ADR, ADDRESS;
   IMPORT GEMAESbase;
   IMPORT GEMDOS;
   IMPORT AESGraphics;
   IMPORT M2Conversions;
   IMPORT Resource;
   IMPORT Screen;
   IMPORT Text;
   IMPORT Icon;
   IMPORT Dialog;
   IMPORT Window;
   IMPORT AppointmentBook;


   CONST Bell = 07C;

   TYPE DateType =
      RECORD
         DateIndex   : INTEGER;
         Appointment : BOOLEAN;   
      END;

   TYPE
      AppointmentType = ARRAY [1..31] OF BOOLEAN;
      String9         = ARRAY [0..9] OF CHAR;

   VAR
      TodaysDate       : CARDINAL;
      TodaysMonth      : CARDINAL;
      TodaysYear       : CARDINAL;

      DisplayedMonth   : CARDINAL;
      DisplayedYear    : CARDINAL;

      DisplayedMonthStartsOn : CARDINAL;
      DaysInDisplayedMonth   : CARDINAL;

      ShowAppointments : BOOLEAN;

      Dates : ARRAY [1..42] OF DateType;

      MonthSum    : ARRAY [1..12] OF CARDINAL;
      MonthLength : ARRAY [1..12] OF CARDINAL;
      MonthName   : ARRAY [1..12] OF String9;

      WindowInfo      : Window.Information;
      WindowAllocated : BOOLEAN;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DayOfWeek (Day, Month, Year : CARDINAL) : CARDINAL;
   VAR
      TotalDays : CARDINAL;   
      LeapYear  : CARDINAL;  
 
   BEGIN
      LeapYear := 0;
      IF (Month>2) OR ((Month=2) AND (Day=29)) THEN
        IF (Year MOD 400)=0 THEN
          LeapYear := 1;
        ELSIF (Year MOD 100)=0 THEN
          LeapYear := 0;
        ELSIF (Year MOD 4)=0 THEN
          LeapYear := 1;
        END;
      END;

      TotalDays := (Year-1)  + (Year-1) DIV 4;
      TotalDays := TotalDays - (Year-1) DIV 100;
      TotalDays := TotalDays + (Year-1) DIV 400;
      
      TotalDays := TotalDays + MonthSum[Month] + Day + LeapYear;

      RETURN ((TotalDays MOD 7) + 1);
   END DayOfWeek;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DaysInMonth (Month, Year : CARDINAL) : CARDINAL;
   VAR
      LeapYear : CARDINAL;

   BEGIN
      LeapYear := 0;
      IF Month=2 THEN
        IF (Year MOD 400)=0 THEN
          LeapYear := 1;
        ELSIF (Year MOD 100)=0 THEN
          LeapYear := 0;
        ELSIF (Year MOD 4)=0 THEN
          LeapYear := 1;
        END;
      END;

      RETURN (MonthLength[Month] + LeapYear);
   END DaysInMonth;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE MapStateToObjects ();

   TYPE TedInfoPtrType = POINTER TO Icon.Tedinfo;    

   CONST

      (* The next three constants define bit patterns determining *)
      (* the color and background pattern for BOXTEXT objects :   *)
      (*                                                          *)
      (* Border color = 4 bits                                    *)
      (* Text color   = 4 bits                                    *)
      (* Text Mode    = 1 bit, 0 = transparent, 1 = replace       *)
      (* Fill pattern = 3 bits                                    *)
      (* Inside color = 4 bits                                    *)

      HighLighted = 0141H;
      Colored     = 0280H; 
      Normal      = 0180H;

   VAR
     
      I             : CARDINAL;
      ObjectAddress : Icon.ObjectPtr;
      TextPtr       : Text.StringPtr;
      DisplayedDate : CARDINAL;
      InsideColor   : LONGCARD;
      TedinfoPtr    : TedInfoPtrType;
   

   BEGIN

      (* Set up the resource objects according to the       *)
      (* system state variables.                            *)

      (* Set the month and year objects... *)
      ObjectAddress := Icon.GetAddress(
         Resource.D10CALEN,
         Resource.D10MNTH1);
      ObjectAddress^.Spec := ADR (MonthName[DisplayedMonth]);

      ObjectAddress := Icon.GetAddress(
         Resource.D10CALEN,
         Resource.D10YEAR1);
      TextPtr := Text.StringPtr (ObjectAddress^.Spec);
      M2Conversions.ConvertCardinal ( DisplayedYear, 4, TextPtr^ );   


      (* Set all dates' fields string values to blanks, and *) 
      (* their states to un-selected and un-outlined.       *) 

      FOR I := 1 TO 42 DO
         ObjectAddress := Icon.GetAddress(
            Resource.D10CALEN,
            Dates[I].DateIndex);
         TedinfoPtr := Text.StringPtr (ObjectAddress^.Spec);
         TextPtr    := Text.StringPtr(TedinfoPtr^.Text);
         Text.Assign ( "  ",TextPtr^ );
         TedinfoPtr^.Color := Normal;
         ObjectAddress^.State := ObjectAddress^.State - {Icon.Selected};
      END;

      (* Set the dates' fields which correspond to days in the  *)
      (* displayed month to the correct value, outlining todays *)
      (* date, and selecting those dates with appointments.     *)

      FOR I := DisplayedMonthStartsOn TO
         (DisplayedMonthStartsOn + DaysInDisplayedMonth - 1) DO 
         
         DisplayedDate :=  I - DisplayedMonthStartsOn + 1;
         ObjectAddress := Icon.GetAddress(
            Resource.D10CALEN, Dates[I].DateIndex);

         TedinfoPtr := Text.StringPtr (ObjectAddress^.Spec);
         TextPtr    := Text.StringPtr(TedinfoPtr^.Text);
         M2Conversions.ConvertInteger ( DisplayedDate, 2, TextPtr^ );   

         IF (DisplayedYear  = TodaysYear)  AND
            (DisplayedMonth = TodaysMonth) AND
            (DisplayedDate  = TodaysDate)  THEN
              ObjectAddress^.State := ObjectAddress^.State + {Icon.Selected};
         END;

         IF ShowAppointments AND Dates[I].Appointment THEN
            ObjectAddress := Icon.GetAddress(
               Resource.D10CALEN,
               Dates[I].DateIndex);
            TedinfoPtr := TedInfoPtrType (ObjectAddress^.Spec);

            IF Screen.Resolution = Screen.HighRes THEN
               TedinfoPtr^.Color := HighLighted;
            ELSE
               TedinfoPtr^.Color := Colored;
            END;

         END;
      END;
      Window.MoveHorizSlider (
         WindowInfo.Id,
         (1000 * DisplayedMonth - 1000) DIV 11 );

   END MapStateToObjects;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE SetMonthYear ( NewMonth, NewYear : CARDINAL );
   VAR
      I         : CARDINAL;
      DateArray : AppointmentBook.CalenderType;

   BEGIN
      (* Wipe out the old month's appointments. *)
      FOR I := 1 TO 42 DO
         Dates[I].Appointment := FALSE;
      END;

      (* Find out what day of the week this month starts on.    *)

      DisplayedMonthStartsOn := DayOfWeek(1, NewMonth, NewYear);

      (* Find out how many days there are this month.           *)

      DaysInDisplayedMonth := DaysInMonth(NewMonth, NewYear);

      IF ShowAppointments THEN
         (* Get the appointments for the new month from   *)
         (* appointment book, and set the appmnt. fields. *)

         AppointmentBook.DatesWithAppts ( NewMonth, NewYear, DateArray );

         FOR I := DisplayedMonthStartsOn TO
               (DisplayedMonthStartsOn + DaysInDisplayedMonth - 1) DO
            Dates[I].Appointment :=
               DateArray[ I - DisplayedMonthStartsOn + 1 ];
         END;
      END;

      DisplayedYear  := NewYear;
      DisplayedMonth := NewMonth;

      (* Map the state variables into the resource objects. *)
      MapStateToObjects ();
      Window.Redraw ( WindowInfo.Id, WindowInfo.Borders ); 
  END SetMonthYear;
 
   (************************** LOCAL ROUTINE ***************************)
 
   PROCEDURE CalendarDraw ( Region : Screen.Box );

   BEGIN
      ;  (* Drawn by the window manager. *)
   END CalendarDraw;

   (********************************************************************)

   PROCEDURE Initialize ();
   VAR  
      I                : CARDINAL;     (* Loop Counter. *)
      Date             : CARDINAL;     (* System date.  *)
 
   BEGIN

      (* Set up an array which records the number of days between *)
      (* January 1 and the month whose number is the array index. *)
      (* Also, set up an array of the number of days per month.   *)

      MonthSum[ 1] :=   0;   MonthLength[ 1] := 31;
      MonthSum[ 2] :=  31;   MonthLength[ 2] := 28;
      MonthSum[ 3] :=  59;   MonthLength[ 3] := 31;
      MonthSum[ 4] :=  90;   MonthLength[ 4] := 30;
      MonthSum[ 5] := 120;   MonthLength[ 5] := 31;
      MonthSum[ 6] := 151;   MonthLength[ 6] := 30;
      MonthSum[ 7] := 181;   MonthLength[ 7] := 31;
      MonthSum[ 8] := 212;   MonthLength[ 8] := 31;
      MonthSum[ 9] := 243;   MonthLength[ 9] := 30;
      MonthSum[10] := 273;   MonthLength[10] := 31;
      MonthSum[11] := 304;   MonthLength[11] := 30;
      MonthSum[12] := 334;   MonthLength[12] := 31;

      MonthName[1]  := "January  ";
      MonthName[2]  := "February ";
      MonthName[3]  := "March    ";
      MonthName[4]  := "April    ";
      MonthName[5]  := "May      ";
      MonthName[6]  := "June     ";
      MonthName[7]  := "July     ";
      MonthName[8]  := "August   ";
      MonthName[9]  := "September";
      MonthName[10] := "October  ";
      MonthName[11] := "November ";
      MonthName[12] := "December ";

      (* Set up the the array which maps from logical Resource    *)
      (* Object Dates to the actual Resource Object identifier    *)
      (* numbers.  This is only done once**when the system boots. *)

      Dates[ 1].DateIndex := Resource.D10DAY1;
      Dates[ 2].DateIndex := Resource.D10DAY2;
      Dates[ 3].DateIndex := Resource.D10DAY3;
      Dates[ 4].DateIndex := Resource.D10DAY4;
      Dates[ 5].DateIndex := Resource.D10DAY5;
      Dates[ 6].DateIndex := Resource.D10DAY6;
      Dates[ 7].DateIndex := Resource.D10DAY7;
      Dates[ 8].DateIndex := Resource.D10DAY8;
      Dates[ 9].DateIndex := Resource.D10DAY9;
      Dates[10].DateIndex := Resource.D10DAY10;
      Dates[11].DateIndex := Resource.D10DAY11;
      Dates[12].DateIndex := Resource.D10DAY12;
      Dates[13].DateIndex := Resource.D10DAY13;
      Dates[14].DateIndex := Resource.D10DAY14;
      Dates[15].DateIndex := Resource.D10DAY15;
      Dates[16].DateIndex := Resource.D10DAY16;
      Dates[17].DateIndex := Resource.D10DAY17;
      Dates[18].DateIndex := Resource.D10DAY18;
      Dates[19].DateIndex := Resource.D10DAY19;
      Dates[20].DateIndex := Resource.D10DAY20;
      Dates[21].DateIndex := Resource.D10DAY21;
      Dates[22].DateIndex := Resource.D10DAY22;
      Dates[23].DateIndex := Resource.D10DAY23;
      Dates[24].DateIndex := Resource.D10DAY24;
      Dates[25].DateIndex := Resource.D10DAY25;
      Dates[26].DateIndex := Resource.D10DAY26;
      Dates[27].DateIndex := Resource.D10DAY27;
      Dates[28].DateIndex := Resource.D10DAY28;
      Dates[29].DateIndex := Resource.D10DAY29;
      Dates[30].DateIndex := Resource.D10DAY30;
      Dates[31].DateIndex := Resource.D10DAY31;
      Dates[32].DateIndex := Resource.D10DAY32;
      Dates[33].DateIndex := Resource.D10DAY33;
      Dates[34].DateIndex := Resource.D10DAY34;
      Dates[35].DateIndex := Resource.D10DAY35;
      Dates[36].DateIndex := Resource.D10DAY36;
      Dates[37].DateIndex := Resource.D10DAY37;
      Dates[38].DateIndex := Resource.D10DAY38;
      Dates[39].DateIndex := Resource.D10DAY39;
      Dates[40].DateIndex := Resource.D10DAY40;
      Dates[41].DateIndex := Resource.D10DAY41;
      Dates[42].DateIndex := Resource.D10DAY42; 

      (* Set up the state variables for the Calendar.           *)
 
      (* Set the state of the system to show the appointments.  *) 
 
      ShowAppointments := FALSE;   

      (* Set Todays Date to the system clock date. *)

      GEMDOS.GetDate (Date);
      TodaysYear  := (Date DIV 512) + 1980;
      TodaysMonth := (Date MOD 512) DIV 32;
      TodaysDate  := Date MOD 32;

      (* Set Displayed Date to 0/0, so that appointments will   *)
      (* be displayed by SetMonthYear.  SetMonthYear does not   *)
      (* get appointments if the DisplayedDate and the New Date *)
      (* are the same.  This prevents unneeded disk accesses.   *)
      (* By setting the Displayed Date to 0/0 it will never     *)
      (* match Todays Date, unless the user is a CRAZED IDIOT.  *)

      DisplayedYear  := 0;
      DisplayedMonth := 0;

      SetMonthYear(TodaysMonth, TodaysYear);

      (* All state variables have been set up. *)
    
   END Initialize;

   (********************************************************************)

   PROCEDURE DayIsSaturday () : BOOLEAN;
   BEGIN
      RETURN (DayOfWeek(TodaysDate, TodaysMonth, TodaysYear) = 7);
   END DayIsSaturday;

   (********************************************************************)

   PROCEDURE DayIsSunday () : BOOLEAN;
   BEGIN
      RETURN (DayOfWeek(TodaysDate, TodaysMonth, TodaysYear) = 1);
   END DayIsSunday;

   (********************************************************************)

   PROCEDURE OwnsWindow ( WindowId : INTEGER ) : BOOLEAN;

   BEGIN
      IF WindowAllocated THEN
         RETURN (WindowId = WindowInfo.Id);
      ELSE
         RETURN (FALSE);
      END;
   END OwnsWindow;

   (********************************************************************)

   PROCEDURE Open ();
   
   VAR
      WindowRegion      : Screen.Box;
      VirtualRegionSize : Window.Area;

   BEGIN
      IF WindowAllocated THEN
         (* Window already exists, make it the top window *)
         Window.Top ( WindowInfo.Id );
      ELSE
         Icon.GetRegion (Resource.D10CALEN, 0, WindowRegion);
         WindowRegion.Origin.X := 
            Screen.Center.Origin.X - (WindowRegion.Size.Width DIV 2);
         WindowRegion.Origin.Y := 
            Screen.Center.Origin.Y - (WindowRegion.Size.Height DIV 2);
         VirtualRegionSize.Width :=
            Window.Pixel (WindowRegion.Size.Width) * 12;
         VirtualRegionSize.Height :=
            Window.Pixel (WindowRegion.Size.Height);
         IF Window.Open (
               Screen.Center,
               Window.InformationPtr (ADR (WindowInfo)),
               " Calendar ",
               "",    
               WindowRegion,
               VirtualRegionSize,
               GEMAESbase.Name +
               GEMAESbase.Closer +
               GEMAESbase.Mover +
               GEMAESbase.UpArrow +
               GEMAESbase.DownArrow +
               GEMAESbase.LeftArrow +
               GEMAESbase.RightArrow +
               GEMAESbase.HorizSlide,
               0, 1,
               Resource.D10CALEN,
               CalendarDraw ) THEN
            WindowAllocated := TRUE;
            (* Take the current system state and map it *)
            (* to the resource objects.                 *)
            MapStateToObjects ();
         ELSE
            Window.Unavailable;
            WindowAllocated := FALSE;
         END;
      END;
   END Open;

   (********************************************************************)

   PROCEDURE Close ();

   BEGIN
      IF WindowAllocated THEN
         Window.Close ( WindowInfo.Id, Screen.Center );
         WindowAllocated := FALSE;
      END;
   END Close;

   (********************************************************************)

   PROCEDURE ProcessMessageEvent ( VAR MsgBuffer : ARRAY OF INTEGER ); 

   CONST
      RowUp       = 2;
      RowDown     = 3;
      PageLeft    = 4;
      PageRight   = 5;
      ColumnLeft  = 6;
      ColumnRight = 7;
      AddAppt     = TRUE;
      DeleteAppt  = FALSE;
      
   VAR
      NewMonth       : CARDINAL;
      NewYear        : CARDINAL;
      MonthIsCurrent : BOOLEAN;

   BEGIN
      IF MsgBuffer[0] = GEMAESbase.AccessoryClose THEN
         IF WindowAllocated THEN
            Window.Close ( 0, Screen.Center );
            WindowAllocated := FALSE;
         END;

      ELSIF MsgBuffer[0] = GEMAESbase.WindowArrowed THEN
         AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );

         IF MsgBuffer[4] = RowUp THEN
            (* Go FORWARD in time by one YEAR. *)
            SetMonthYear(DisplayedMonth, DisplayedYear+1);

         ELSIF MsgBuffer[4] = RowDown THEN
            (* Go BACKWARD in time by one YEAR. *)
            SetMonthYear(DisplayedMonth, DisplayedYear-1);

         ELSIF (MsgBuffer[4] = PageLeft) OR
               (MsgBuffer[4] = ColumnLeft) THEN
            (* Go BACKWARD in time by one MONTH. *)
            IF DisplayedMonth = 1 THEN
               NewMonth := 12;
               NewYear  := DisplayedYear-1;
            ELSE
               NewMonth := DisplayedMonth-1;
               NewYear  := DisplayedYear;
            END; 
            SetMonthYear(NewMonth, NewYear);

         ELSIF (MsgBuffer[4] = PageRight) OR
               (MsgBuffer[4] = ColumnRight) THEN
            (* Go FORWARD in time by one MONTH. *);
            IF DisplayedMonth = 12 THEN
               NewMonth := 1;
               NewYear  := DisplayedYear+1;
            ELSE
               NewMonth := DisplayedMonth+1;
               NewYear  := DisplayedYear;
            END; 
            SetMonthYear(NewMonth, NewYear);

         END;
         AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );

      ELSIF MsgBuffer[0] = GEMAESbase.WindowHorizSlided THEN
         AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );

         (* Value in MsgBuffer[4] is from 0 to 1000, and    *)
         (* represents the position of the slide box inside *)
         (* the scroll region.  Go to the month which is    *)
         (* closest to the position of the slide box.       *)

         NewMonth := (11 * MsgBuffer[4] + 1000) DIV 1000;
         SetMonthYear(NewMonth, DisplayedYear);
         AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );

      ELSIF MsgBuffer[0] = Resource.DateChanged THEN

         (* Update the state variables. *)

         MonthIsCurrent :=
            (DisplayedYear  = TodaysYear)  AND
            (DisplayedMonth = TodaysMonth);
         TodaysYear  := MsgBuffer[5]; 
         TodaysMonth := MsgBuffer[4]; 
         TodaysDate  := MsgBuffer[3];

         IF (DisplayedYear  = TodaysYear)  AND
            (DisplayedMonth = TodaysMonth) THEN
            (* Move the outline to the correct date. *)
            MapStateToObjects ();
            Window.Redraw ( WindowInfo.Id, WindowInfo.Borders ); 
         ELSIF MonthIsCurrent THEN
            SetMonthYear ( TodaysMonth, TodaysYear );
         END;

      ELSIF (MsgBuffer[0] = Resource.AddedAppointment) OR
            (MsgBuffer[0] = Resource.DeletedAppointment) THEN
         IF WindowAllocated AND ShowAppointments THEN
            SetMonthYear ( DisplayedMonth, DisplayedYear );
         END;

      ELSE
         GEMDOS.ConOut ( Bell );
      END;
   END ProcessMessageEvent;

   (********************************************************************)

   PROCEDURE ProcessButtonEvent ( MouseButton : INTEGER;
                                  ModifierKey : INTEGER;
                                  Clicks      : INTEGER );

   VAR
      MouseLocation : Screen.PixelCoordinate;
      MouseState    : INTEGER;
      KeyboardState : INTEGER;
      Object        : INTEGER;
      ObjectFound   : BOOLEAN;
      ObjectPtr     : Icon.ObjectPtr;
      TextPtr       : Text.StringPtr;
      I             : CARDINAL;
      Date          : CARDINAL;
   
   BEGIN
      AESGraphics.GrafMouseKeyboardState (
         MouseLocation.X,
         MouseLocation.Y,
         MouseState,
         KeyboardState );
      IF Icon.Find ( Resource.D10CALEN, MouseLocation, Object ) THEN
         ObjectPtr := Icon.GetAddress ( Resource.D10CALEN, Object );
         IF (ObjectPtr^.Type = GEMAESbase.GraphicBoxText) AND (Clicks = 2) THEN
            ObjectFound := FALSE;
            I := 1;
            LOOP
               IF I > 42 THEN 
                 EXIT;
               END; 
               IF (Object = Dates[I].DateIndex) AND 
                  (I >= DisplayedMonthStartsOn) AND
                  (I <= (DisplayedMonthStartsOn+DaysInDisplayedMonth-1))
               THEN
                  ObjectFound := TRUE;
                  AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );
                  IF (DisplayedYear  = TodaysYear ) AND
                     (DisplayedMonth = TodaysMonth) AND
                     (I-DisplayedMonthStartsOn+1 = TodaysDate) 
                  THEN
                     (* Appmnt. book opened on today. *)
                     AppointmentBook.Open ( TRUE, 0 );
                  ELSE
                     (* Appmnt. book opened on day other than today. *)
                     Date :=
                        ((DisplayedYear - 1980) * 512) +
                        (DisplayedMonth * 32) +
                        (I-DisplayedMonthStartsOn+1);
                     AppointmentBook.Open ( FALSE, Date );
                  END;
                  EXIT;
               END; (* IF *)
               INC(I);
            END (* LOOP *);

            IF NOT ObjectFound THEN
              GEMDOS.ConOut(Bell);
            END;

         ELSIF Object = Resource.D10APPT THEN
            AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );
            ObjectPtr^.State := ObjectPtr^.State + {Icon.Selected};
            Icon.Display (
               Resource.D10CALEN,
               Resource.D10APPT,
               WindowInfo.Borders );
            ObjectPtr^.State := ObjectPtr^.State - {Icon.Selected};
            TextPtr := Text.StringPtr (ObjectPtr^.Spec);
            IF ShowAppointments THEN
               ShowAppointments := FALSE;
               TextPtr^ := "Show Appointments";
               MapStateToObjects;
               Window.Redraw ( WindowInfo.Id, WindowInfo.Borders ); 
            ELSE
               ShowAppointments := TRUE;
               TextPtr^ := "Hide Appointments";
               SetMonthYear ( DisplayedMonth, DisplayedYear );
            END;

         ELSE
            GEMDOS.ConOut ( Bell );
         END;
      END;         
      AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
   END ProcessButtonEvent;

   (********************************************************************)

   PROCEDURE ProcessKeyboardEvent ( Key         : INTEGER;
                                    ModifierKey : INTEGER );

   BEGIN
      GEMDOS.ConOut ( Bell );
   END ProcessKeyboardEvent;


BEGIN (* Module initialization code. *)
   WindowAllocated := FALSE;
END Calendar.

