IMPLEMENTATION MODULE Event;

   (* This module has been modified specifically for the DeskPak *)
   (* desk accessories application.                              *)

   (*$S-,$T- turn off stack and range checking *) 

   FROM SYSTEM IMPORT ADR, ADDRESS;
   IMPORT GEMAESbase;
   IMPORT AESObjects;
   IMPORT Screen;
   IMPORT Resource;
   IMPORT Icon;
   IMPORT Dialog;
   IMPORT Window;
   IMPORT File;
   IMPORT Clock;
   IMPORT Calendar;
   IMPORT PhoneBook;
   IMPORT Calculator;
   IMPORT AppointmentBook;
   IMPORT NotePad;
   IMPORT FreeRam;

   (********************************************************************)

   PROCEDURE Initialize ();

   BEGIN
      Clock.Initialize;
      Calculator.Initialize;
      Calendar.Initialize;
      PhoneBook.Initialize;
      NotePad.Initialize;
      AppointmentBook.Initialize;
   END Initialize;

   (********************************************************************)

   PROCEDURE ProcessTimer;

   BEGIN
      Clock.ProcessTimerEvent ();
   END ProcessTimer;
        
   (********************************************************************)

   PROCEDURE ProcessMessage ( WindowId      : INTEGER;
                              VAR MsgBuffer : ARRAY OF INTEGER);

   VAR
      Region     : Screen.Box;
      ExitObject : INTEGER;

   BEGIN
      IF MsgBuffer[0] = GEMAESbase.WindowRedraw THEN
         Region.Origin.X    := MsgBuffer[4];
         Region.Origin.Y    := MsgBuffer[5];
         Region.Size.Width  := MsgBuffer[6];
         Region.Size.Height := MsgBuffer[7];
         Window.Redraw ( WindowId, Region ); 

      ELSIF MsgBuffer[0] = GEMAESbase.WindowTopped THEN
         Window.Top ( WindowId );

      ELSIF MsgBuffer[0] = GEMAESbase.WindowClosed THEN
         IF Clock.OwnsWindow ( WindowId ) THEN
            Clock.Close ();
         ELSIF Calculator.OwnsWindow ( WindowId ) THEN
            Calculator.Close ();
         ELSIF PhoneBook.OwnsWindow ( WindowId ) THEN
            PhoneBook.Close ();
         ELSIF Calendar.OwnsWindow ( WindowId ) THEN
            Calendar.Close ();
         ELSIF AppointmentBook.OwnsWindow ( WindowId ) THEN
            AppointmentBook.Close ();
         ELSIF NotePad.OwnsWindow ( WindowId ) THEN
            NotePad.Close ();
         END;

      ELSIF MsgBuffer[0] = GEMAESbase.WindowFulled THEN
         IF Clock.OwnsWindow ( WindowId ) THEN
            Clock.ProcessMessageEvent ( MsgBuffer ); 
         END;

      ELSIF MsgBuffer[0] = GEMAESbase.WindowMoved THEN
         Region.Origin.X := MsgBuffer[4];
         Region.Origin.Y := MsgBuffer[5];
         Window.Move ( WindowId, Region.Origin );

      ELSIF MsgBuffer[0] = GEMAESbase.AccessoryOpen THEN
         IF MsgBuffer[4] = AccessoryMenuId THEN
            LOOP
               ExitObject := Dialog.Display (
                  Resource.D13TOOLS, Screen.Center, Dialog.NoEditableField );
               IF ExitObject = Resource.D13CLOCK THEN
                  Clock.Open ();
                  EXIT;
               ELSIF ExitObject = Resource.D13CALC THEN
                  Calculator.Open ();
                  EXIT;
               ELSIF ExitObject = Resource.D13PHONE THEN
                  PhoneBook.Open ();
                  EXIT;
               ELSIF ExitObject = Resource.D13CALEN THEN
                  Calendar.Open ();
                  EXIT;
               ELSIF ExitObject = Resource.D13APPT THEN
                  AppointmentBook.Open ( TRUE, 0 );   (* Use current date *)
                  EXIT;
               ELSIF ExitObject = Resource.D13FREE THEN
                  FreeRam.Display;
               ELSIF ExitObject = Resource.D13NOTE THEN
                  NotePad.Open ();
                  EXIT;
               ELSIF ExitObject = Resource.D13COPY THEN
                  File.Copy;
                  EXIT;
               ELSIF ExitObject = Resource.D13DEL THEN
                  File.Delete;
                  EXIT;
               ELSIF ExitObject = Resource.D13AE THEN
                  ExitObject := Dialog.Display (
                     Resource.D14INFO, Screen.Center, Dialog.NoEditableField );
               ELSE
                  EXIT;
               END;
            END (* LOOP *);
         END;

      ELSIF MsgBuffer[0] = GEMAESbase.AccessoryClose THEN
         Clock.ProcessMessageEvent ( MsgBuffer ); 
         Calculator.ProcessMessageEvent ( MsgBuffer );
         PhoneBook.ProcessMessageEvent ( MsgBuffer );
         Calendar.ProcessMessageEvent ( MsgBuffer );
         AppointmentBook.ProcessMessageEvent ( MsgBuffer ); 
         NotePad.ProcessMessageEvent ( MsgBuffer );

      ELSIF MsgBuffer[0] = Resource.MinuteChanged THEN
         PhoneBook.ProcessMessageEvent ( MsgBuffer );
         AppointmentBook.ProcessMessageEvent ( MsgBuffer );

      ELSIF (MsgBuffer[0] = Resource.DateChanged) THEN
         Calendar.ProcessMessageEvent ( MsgBuffer );
         AppointmentBook.ProcessMessageEvent ( MsgBuffer );

      ELSIF (MsgBuffer[0] = Resource.AddedAppointment) OR
            (MsgBuffer[0] = Resource.DeletedAppointment) THEN
         Calendar.ProcessMessageEvent ( MsgBuffer );

      ELSIF MsgBuffer[0] = Resource.CallBack THEN
         AppointmentBook.ProcessMessageEvent ( MsgBuffer );

      ELSIF Clock.OwnsWindow ( WindowId ) THEN
         Clock.ProcessMessageEvent ( MsgBuffer ); 

      ELSIF Calculator.OwnsWindow ( WindowId ) THEN
         Calculator.ProcessMessageEvent ( MsgBuffer );

      ELSIF PhoneBook.OwnsWindow ( WindowId ) THEN
         PhoneBook.ProcessMessageEvent ( MsgBuffer );

      ELSIF Calendar.OwnsWindow ( WindowId ) THEN
         Calendar.ProcessMessageEvent ( MsgBuffer );

      ELSIF AppointmentBook.OwnsWindow ( WindowId ) THEN
         AppointmentBook.ProcessMessageEvent ( MsgBuffer );

      ELSIF NotePad.OwnsWindow ( WindowId ) THEN
         NotePad.ProcessMessageEvent ( MsgBuffer );

      END;
   END ProcessMessage;

   (********************************************************************)

   PROCEDURE ProcessButton ( WindowId    : INTEGER;
                             MouseButton : INTEGER;
                             ModifierKey : INTEGER;
                             Clicks      : INTEGER );

   BEGIN
      IF Clock.OwnsWindow ( WindowId ) THEN
         Clock.ProcessButtonEvent (
            MouseButton, ModifierKey, Clicks );
      ELSIF Calculator.OwnsWindow ( WindowId ) THEN
         Calculator.ProcessButtonEvent (
            MouseButton, ModifierKey, Clicks );
      ELSIF PhoneBook.OwnsWindow ( WindowId ) THEN
         PhoneBook.ProcessButtonEvent (
            MouseButton, ModifierKey, Clicks );
      ELSIF Calendar.OwnsWindow ( WindowId ) THEN
         Calendar.ProcessButtonEvent (
            MouseButton, ModifierKey, Clicks );
      ELSIF AppointmentBook.OwnsWindow ( WindowId ) THEN
         AppointmentBook.ProcessButtonEvent (
            MouseButton, ModifierKey, Clicks );
      ELSIF NotePad.OwnsWindow ( WindowId ) THEN
         NotePad.ProcessButtonEvent (
            MouseButton, ModifierKey, Clicks );
      END;
   END ProcessButton;

   (********************************************************************)

   PROCEDURE ProcessKey ( WindowId    : INTEGER;
                          Key         : INTEGER;
                          ModifierKey : INTEGER );

   BEGIN
      IF Clock.OwnsWindow ( WindowId ) THEN
         Clock.ProcessKeyboardEvent ( Key, ModifierKey );
      ELSIF Calculator.OwnsWindow ( WindowId ) THEN
         Calculator.ProcessKeyboardEvent ( Key, ModifierKey );
      ELSIF PhoneBook.OwnsWindow ( WindowId ) THEN
         PhoneBook.ProcessKeyboardEvent ( Key, ModifierKey );
      ELSIF Calendar.OwnsWindow ( WindowId ) THEN
         Calendar.ProcessKeyboardEvent ( Key, ModifierKey );
      ELSIF AppointmentBook.OwnsWindow ( WindowId ) THEN
         AppointmentBook.ProcessKeyboardEvent ( Key, ModifierKey );
      ELSIF NotePad.OwnsWindow ( WindowId ) THEN
         NotePad.ProcessKeyboardEvent ( Key, ModifierKey );
      END;
   END ProcessKey;

END Event.

