IMPLEMENTATION MODULE Float;


   (*$S-,$T- turn off stack and range checking *) 

   IMPORT MathLib0;
   IMPORT LongMathLib0;
   IMPORT Conversions;
   IMPORT Readout;


   TYPE IntegerBits = SET OF [0..31];

   VAR
      NumberBase    : CARDINAL;
      LongZero      : LONGREAL;
      LongOne       : LONGREAL;
      Long7FFFFFFFH : LONGREAL;

   (*--------------------------------------------------------------------*)

   PROCEDURE SetBase (
      Base       : (* IN *) CARDINAL;
      VAR Number : (* IN OUT *) DataElement ) : BOOLEAN;

   VAR
      Success  : BOOLEAN;
      Negative : BOOLEAN;

   BEGIN
      Success := TRUE;
      IF (Base = 10) AND (NumberBase <> 10) THEN
         Negative := (Number.IntegerItem < 0);
         Number.IntegerItem := ABS (Number.IntegerItem);
         Number.RealItem    := FLOATD (LONGCARD (Number.IntegerItem));
         IF Negative THEN
            Number.RealItem := -Number.RealItem;
         END;
      ELSIF (Base <> 10) AND (NumberBase = 10) THEN
         Negative := (Number.RealItem < LongZero);
         Number.RealItem := ABS (Number.RealItem);
         IF Number.RealItem > Long7FFFFFFFH THEN
            Number.IntegerItem := 0;
            Success := FALSE;
         ELSE
            Number.IntegerItem := LONGINT (TRUNCD (Number.RealItem));
            IF Negative THEN
               Number.IntegerItem := -Number.IntegerItem;
            END;
         END;
      END;
      NumberBase := Base;
      RETURN (Success);
   END SetBase;

   (*--------------------------------------------------------------------*)

   PROCEDURE Compare ( Number (* IN *) : DataElement ) : CompareResult;

   BEGIN
      IF NumberBase = 10 THEN
         IF Number.RealItem < LongZero THEN
            RETURN (Negative);
         ELSIF Number.RealItem = LongZero THEN
            RETURN (Zero);
         ELSE
            RETURN (Positive);
         END;
      ELSE
         IF Number.IntegerItem < 0 THEN
            RETURN (Negative);
         ELSIF Number.IntegerItem = 0 THEN
            RETURN (Zero);
         ELSE 
            RETURN (Positive);
         END;
      END;
   END Compare;

   (*--------------------------------------------------------------------*)

   PROCEDURE MakeSmallLiteral (
      Literal    : (* IN *)  INTEGER;
      VAR Result : (* OUT *) DataElement );

   VAR Negative : BOOLEAN;

   BEGIN
      Negative := (Literal < 0);
      Result.IntegerItem := LONGINT (Literal);
      Result.RealItem := FLOATD (CARDINAL (ABS (Literal)));
      IF Negative THEN
         Result.RealItem := -Result.RealItem;
      END;
   END MakeSmallLiteral;

   (*------------------------------------------------------------------*)

   PROCEDURE Image (
      Source     : (* IN *)  DataElement;
      VAR String : (* OUT *) ARRAY OF CHAR );

   VAR
      Width   : INTEGER;
      Success : BOOLEAN;
      Temp    : DataElement;

   BEGIN
      Temp := Source;
      IF NumberBase = 10 THEN
         Conversions.ConvertLongRealToString (
            Temp.RealItem, String, Readout.DisplayLength - 2, Success );
      ELSE
         IF NumberBase = 2 THEN

            (* Special kludge to fit binary numbers into 16 digit *)
            (* readout.  Binary radix will only display the last  *)
            (* 16 digits of the number.                           *)

            Temp.IntegerItem :=
               LONGINT (LONGCARD (Temp.IntegerItem) MOD 10000H);
         END;
         Conversions.ConvertToString (
            LONGCARD (Temp.IntegerItem),
            NumberBase,
            FALSE,
            String,
            Success );
      END;
   END Image;

   (*------------------------------------------------------------------*)

   PROCEDURE Value (
      VAR String      : (* IN *)  ARRAY OF CHAR;
      VAR Destination : (* OUT *) DataElement ) : BOOLEAN;

   VAR
      Success : BOOLEAN;
      Result  : LONGCARD;

   BEGIN
      IF NumberBase = 10 THEN
         Conversions.ConvertLongRealFromString (
            Destination.RealItem,
            String,
            Success );
      ELSE
         Conversions.ConvertFromString (
            String,
            NumberBase,
            TRUE,
            0FFFFFFFFH,
            Result,
            Success );
         Destination.IntegerItem := LONGINT (Result);
      END;
      RETURN (Success);
   END Value;

   (*------------------------------------------------------------------*)

   PROCEDURE ChangeSign (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := -Source.RealItem;
      ELSE
         Result.IntegerItem := -Source.IntegerItem;
      END;
   END ChangeSign;

   (*------------------------------------------------------------------*)

   PROCEDURE OnesComplement (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   VAR
      SourceBits   : IntegerBits;
      FullSet      : IntegerBits;

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Error := TRUE;
      ELSE 
         FullSet    := IntegerBits (0FFFFFFFFH);
         SourceBits := IntegerBits (Source.IntegerItem);
         Result.IntegerItem := LONGINT (SourceBits / FullSet);
      END;
   END OnesComplement;

   (*------------------------------------------------------------------*)

   PROCEDURE Add (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := Left.RealItem + Right.RealItem;
      ELSE
         Result.IntegerItem := Left.IntegerItem + Right.IntegerItem;
      END;
   END Add;

   (*------------------------------------------------------------------*)

   PROCEDURE Subtract (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := Left.RealItem - Right.RealItem;
      ELSE
         Result.IntegerItem := Left.IntegerItem - Right.IntegerItem;
      END;
   END Subtract;

   (*------------------------------------------------------------------*)

   PROCEDURE Multiply (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := Left.RealItem * Right.RealItem;
      ELSE
         Result.IntegerItem := Left.IntegerItem * Right.IntegerItem;
      END;
   END Multiply;

   (*------------------------------------------------------------------*)

   PROCEDURE Divide (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := Left.RealItem / Right.RealItem;
      ELSE
         Result.IntegerItem := Left.IntegerItem DIV Right.IntegerItem;
      END;
   END Divide;

   (*------------------------------------------------------------------*)

   PROCEDURE Sin (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         IF Source.RealItem = LongZero THEN
            Result.RealItem := 0.0;
         ELSE
            Result.RealItem := LongMathLib0.dsin ( Source.RealItem );
         END;
      ELSE
         Error := TRUE;
      END;
   END Sin;

   (*------------------------------------------------------------------*)

   PROCEDURE Cos (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := LongMathLib0.dcos ( Source.RealItem );
      ELSE
         Error := TRUE;
      END;
   END Cos;

   (*------------------------------------------------------------------*)

   PROCEDURE Tan (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := LongMathLib0.dtan ( Source.RealItem );
      ELSE
         Error := TRUE;
      END;
   END Tan;

   (*------------------------------------------------------------------*)

   PROCEDURE ASin (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         IF ABS (Source.RealItem) > LongOne THEN
            Error := TRUE;
         ELSIF ABS (Source.RealItem) = LongOne THEN
            Result.RealItem := LONG (MathLib0.pi) / 2.0;
         ELSE
            Result.RealItem := LongMathLib0.darcsin ( Source.RealItem );
         END;
      ELSE
         Error := TRUE;
      END;
   END ASin;

   (*------------------------------------------------------------------*)

   PROCEDURE ACos (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         IF ABS (Source.RealItem) > LongOne THEN
            Error := TRUE;
         ELSIF ABS (Source.RealItem) = LongOne THEN
            Result.RealItem := 0.0;
         ELSE
            Result.RealItem := LongMathLib0.darccos ( Source.RealItem );
         END;
      ELSE
         Error := TRUE;
      END;
   END ACos;

   (*------------------------------------------------------------------*)

   PROCEDURE ATan (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := LongMathLib0.darctan ( Source.RealItem );
      ELSE
         Error := TRUE;
      END;
   END ATan;

   (*------------------------------------------------------------------*)

   PROCEDURE Reciprocal (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         IF Source.RealItem = LongZero THEN
            Error := TRUE;
         ELSE
            Result.RealItem := LongOne / Source.RealItem;
         END;
      ELSE
         Error := TRUE;
      END;
   END Reciprocal;

   (*------------------------------------------------------------------*)

   PROCEDURE SquareRoot (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         IF Source.RealItem < LongZero THEN
            Error := TRUE;
         ELSE
            Result.RealItem := LongMathLib0.dsqrt ( Source.RealItem );
         END;
      ELSE
         Error := TRUE;
      END;
   END SquareRoot;

   (*------------------------------------------------------------------*)

   PROCEDURE Power (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Result.RealItem := LongMathLib0.dpower ( Left.RealItem, Right.RealItem );
      ELSE
         Error := TRUE;
      END;
   END Power;

   (*------------------------------------------------------------------*)

   PROCEDURE Log (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         IF Source.RealItem <= LongZero THEN
            Error := TRUE;
         ELSE
            Result.RealItem := LongMathLib0.dlog ( Source.RealItem );
         END;
      ELSE
         Error := TRUE;
      END;
   END Log;

   (*------------------------------------------------------------------*)

   PROCEDURE Ln (
      Source     : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         IF Source.RealItem <= LongZero THEN
            Error := TRUE;
         ELSE
            Result.RealItem := LongMathLib0.dln ( Source.RealItem );
         END;
      ELSE
         Error := TRUE;
      END;
   END Ln;

   (*------------------------------------------------------------------*)

   PROCEDURE Shift (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   VAR Count : LONGINT;

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Error := TRUE;
      ELSIF Right.IntegerItem > 0 THEN
         Count := Right.IntegerItem;
         Result.IntegerItem := Left.IntegerItem;
         LOOP
            Result.IntegerItem := Result.IntegerItem * 2;
            DEC (Count);
            IF Error THEN
               EXIT;
            ELSIF Count = 0 THEN
               EXIT;
            END;
         END (* LOOP *);
      ELSIF Right.IntegerItem < 0 THEN
         Count := -Right.IntegerItem;
         Result.IntegerItem := Left.IntegerItem;
         LOOP
            Result.IntegerItem := Result.IntegerItem DIV 2;
            DEC (Count);
            IF Result.IntegerItem = 0 THEN
               EXIT;
            ELSIF Count = 0 THEN
               EXIT;
            END;
         END (* LOOP *);
      END;
   END Shift;

   (*------------------------------------------------------------------*)

   PROCEDURE And (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   VAR
      LeftBits     : IntegerBits;
      RightBits    : IntegerBits;

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Error := TRUE;
      ELSE 
         LeftBits  := IntegerBits (Left.IntegerItem);
         RightBits := IntegerBits (Right.IntegerItem);
         Result.IntegerItem := LONGINT (LeftBits * RightBits);
      END;
   END And;

   (*------------------------------------------------------------------*)

   PROCEDURE Or (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   VAR
      LeftBits     : IntegerBits;
      RightBits    : IntegerBits;

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Error := TRUE;
      ELSE 
         LeftBits  := IntegerBits (Left.IntegerItem);
         RightBits := IntegerBits (Right.IntegerItem);
         Result.IntegerItem := LONGINT (LeftBits + RightBits);
      END;
   END Or;

   (*------------------------------------------------------------------*)

   PROCEDURE ExclusiveOr (
      Left       : (* IN *)  DataElement;
      Right      : (* IN *)  DataElement;
      VAR Result : (* OUT *) DataElement );

   VAR
      LeftBits     : IntegerBits;
      RightBits    : IntegerBits;

   BEGIN
      Error := FALSE;
      IF NumberBase = 10 THEN
         Error := TRUE;
      ELSE 
         LeftBits  := IntegerBits (Left.IntegerItem);
         RightBits := IntegerBits (Right.IntegerItem);
         Result.IntegerItem := LONGINT (LeftBits / RightBits);
      END;
   END ExclusiveOr;


BEGIN
   NumberBase    := 10;
   LongZero      := 0.0;
   LongOne       := 1.0;
   Long7FFFFFFFH := FLOATD (7FFFFFFFH);
   Error         := FALSE;
END Float.


