IMPLEMENTATION MODULE Keypad;


   (*$S-,$T- turn off stack and range checking *) 

   IMPORT GEMDOS;
   IMPORT Text;
   IMPORT Screen;
   IMPORT Icon;
   IMPORT Resource;
   IMPORT Readout;

   CONST Bell = 07C;

   VAR
      UpdateRegion : Screen.Box;
      DisplayKeys  : BOOLEAN;
      SpecialKey   : INTEGER;         (* Key requiring a parameter *)

      (*--- The following variables define the keypad's state ---------*)

      NumberBase        : INTEGER;
      MaxDigitsForBase  : CARDINAL;
      InProgramMode     : BOOLEAN;
      DigitCount        : CARDINAL;
      PressedCE         : BOOLEAN;
      ProcessedExponent : BOOLEAN;
      Mantissa          : ARRAY [0..Readout.DisplayLength] OF CHAR;
      Exponent          : ARRAY [0..5] OF CHAR;
      KeysDisabled      : BOOLEAN;

      DigitsExpected    : INTEGER;
      DigitBuffer       : ARRAY [0..4] OF CHAR;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE SelectKey ( Key : INTEGER );

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress ( Resource.D9CALC, Key );
      ObjectPtr^.State := ObjectPtr^.State + {Icon.Selected};
      IF DisplayKeys THEN
         Icon.Display ( Resource.D9CALC, Key, UpdateRegion );
      END;
   END SelectKey;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DeselectKey ( Key : INTEGER );

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress ( Resource.D9CALC, Key );
      ObjectPtr^.State := ObjectPtr^.State - {Icon.Selected};
      IF DisplayKeys THEN
         Icon.Display ( Resource.D9CALC, Key, UpdateRegion );
      END;
   END DeselectKey;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE EnableKey ( Key : INTEGER );

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress ( Resource.D9CALC, Key );
      ObjectPtr^.State := ObjectPtr^.State - {Icon.Disabled};
      IF DisplayKeys THEN
         Icon.Display ( Resource.D9CALC, Key, UpdateRegion );
      END;
   END EnableKey;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DisableKey ( Key : INTEGER );

   VAR ObjectPtr : Icon.ObjectPtr;

   BEGIN
      ObjectPtr := Icon.GetAddress ( Resource.D9CALC, Key );
      ObjectPtr^.State := ObjectPtr^.State + {Icon.Disabled};
      IF DisplayKeys THEN
         Icon.Display ( Resource.D9CALC, Key, UpdateRegion );
      END;
   END DisableKey;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE EnableKeysABCDEF;

   BEGIN
      EnableKey ( Resource.D9KEYD );
      EnableKey ( Resource.D9KEYE );
      EnableKey ( Resource.D9KEYF );
      EnableKey ( Resource.D9KEYA );
      EnableKey ( Resource.D9KEYB );
      EnableKey ( Resource.D9KEYC );
   END EnableKeysABCDEF;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DisableKeysABCDEF;

   BEGIN
      DisableKey ( Resource.D9KEYD );
      DisableKey ( Resource.D9KEYE );
      DisableKey ( Resource.D9KEYF );
      DisableKey ( Resource.D9KEYA );
      DisableKey ( Resource.D9KEYB );
      DisableKey ( Resource.D9KEYC );
   END DisableKeysABCDEF;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE EnableKeys89;

   BEGIN
      EnableKey ( Resource.D9KEY8 );
      EnableKey ( Resource.D9KEY9 );
   END EnableKeys89;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DisableKeys89;

   BEGIN
      DisableKey ( Resource.D9KEY8 );
      DisableKey ( Resource.D9KEY9 );
   END DisableKeys89;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE EnableKeys234567;

   BEGIN
      EnableKey ( Resource.D9KEY7 );
      EnableKey ( Resource.D9KEY4 );
      EnableKey ( Resource.D9KEY5 );
      EnableKey ( Resource.D9KEY6 );
      EnableKey ( Resource.D9KEY2 );
      EnableKey ( Resource.D9KEY3 );
   END EnableKeys234567;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DisableKeys234567;

   BEGIN
      DisableKey ( Resource.D9KEY7 );
      DisableKey ( Resource.D9KEY4 );
      DisableKey ( Resource.D9KEY5 );
      DisableKey ( Resource.D9KEY6 );
      DisableKey ( Resource.D9KEY2 );
      DisableKey ( Resource.D9KEY3 );
   END DisableKeys234567;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE ToggleProgramMode;

   BEGIN
      IF InProgramMode THEN
         DeselectKey ( Resource.D9PGM );
         DisableKey ( Resource.D9SST );
         DisableKey ( Resource.D9BST );
         DisableKey ( Resource.D9INS );
         DisableKey ( Resource.D9DEL );
         DisableKey ( Resource.D9GTRZER );
         DisableKey ( Resource.D9GOTO );
         DisableKey ( Resource.D9STOP );
         DisableKey ( Resource.D9SAVE );
         DisableKey ( Resource.D9LOAD );
         DisableKey ( Resource.D9PRINT );
         DisableKey ( Resource.D9EQLZER );
         EnableKey ( Resource.D9RUN );
         InProgramMode := FALSE;
      ELSE
         SelectKey ( Resource.D9PGM );
         EnableKey ( Resource.D9SST );
         EnableKey ( Resource.D9BST );
         EnableKey ( Resource.D9INS );
         EnableKey ( Resource.D9DEL );
         EnableKey ( Resource.D9GTRZER );
         EnableKey ( Resource.D9GOTO );
         EnableKey ( Resource.D9STOP );
         EnableKey ( Resource.D9SAVE );
         EnableKey ( Resource.D9LOAD );
         EnableKey ( Resource.D9PRINT );
         EnableKey ( Resource.D9EQLZER );
         DisableKey ( Resource.D9RUN );
         InProgramMode := TRUE;
      END;
   END ToggleProgramMode;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE EnableDigitKeys ( Base : INTEGER );

   BEGIN
      IF Base = Resource.D9HEX THEN
         EnableKeysABCDEF;
         EnableKeys89;
         EnableKeys234567;
         MaxDigitsForBase := 8;
      ELSIF Base = Resource.D9DEC THEN
         DisableKeysABCDEF;
         EnableKeys89;
         EnableKeys234567;
         MaxDigitsForBase := 11;
      ELSIF Base = Resource.D9OCT THEN
         DisableKeysABCDEF;
         DisableKeys89;
         EnableKeys234567;
         MaxDigitsForBase := 11;
      ELSE (* Base = Resource.D9BIN *)
         DisableKeysABCDEF;
         DisableKeys89;
         DisableKeys234567;
         MaxDigitsForBase := 16;
      END;
   END EnableDigitKeys;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE SetNumberBase ( NewNumberBase : INTEGER );

   BEGIN
      DeselectKey ( Resource.D9DEC );
      DeselectKey ( Resource.D9HEX );
      DeselectKey ( Resource.D9OCT );
      DeselectKey ( Resource.D9BIN );
      SelectKey ( NewNumberBase );
      EnableDigitKeys ( NewNumberBase );
      NumberBase := NewNumberBase;
      IF NewNumberBase = Resource.D9DEC THEN
         EnableKey ( Resource.D9POINT );
         EnableKey ( Resource.D9SIN );
         EnableKey ( Resource.D9COS );
         EnableKey ( Resource.D9TAN );
         EnableKey ( Resource.D9LOG );
         EnableKey ( Resource.D9SQRT );
         EnableKey ( Resource.D9ASIN );
         EnableKey ( Resource.D9ACOS );
         EnableKey ( Resource.D9ATAN );
         EnableKey ( Resource.D9LN );
         EnableKey ( Resource.D9YTOX );
         EnableKey ( Resource.D9RECIP );
         DisableKey ( Resource.D9SHF );
         DisableKey ( Resource.D9AND );
         DisableKey ( Resource.D9OR );
         DisableKey ( Resource.D9XOR );
         DisableKey ( Resource.D91SC );
      ELSE
         DisableKey ( Resource.D9POINT );
         DisableKey ( Resource.D9SIN );
         DisableKey ( Resource.D9COS );
         DisableKey ( Resource.D9TAN );
         DisableKey ( Resource.D9LOG );
         DisableKey ( Resource.D9SQRT );
         DisableKey ( Resource.D9ASIN );
         DisableKey ( Resource.D9ACOS );
         DisableKey ( Resource.D9ATAN );
         DisableKey ( Resource.D9LN );
         DisableKey ( Resource.D9YTOX );
         DisableKey ( Resource.D9RECIP );
         EnableKey ( Resource.D9SHF );
         EnableKey ( Resource.D9AND );
         EnableKey ( Resource.D9OR );
         EnableKey ( Resource.D9XOR );
         EnableKey ( Resource.D91SC );
      END;
   END SetNumberBase;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE ClearDisplay;

   BEGIN
      Readout.Write ( "", UpdateRegion );
      DigitCount        := 0;
      ProcessedExponent := FALSE;
      Mantissa          := " ";
      Exponent          := "E+000";
   END ClearDisplay;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE ConcatChar ( Ch : CHAR );

   VAR
      Buffer  : Text.String80;
      Success : BOOLEAN;

   BEGIN
      IF DigitCount = 0 THEN
         ClearDisplay;
         IF NumberBase = Resource.D9DEC THEN
            EnableKey ( Resource.D9KEYE );
         END;
      END;
      IF (Ch = 'E') AND (NumberBase = Resource.D9DEC) THEN
         ProcessedExponent := TRUE;
         DisableKey ( Resource.D9KEYE );
         DisableKey ( Resource.D9POINT );
         Success := Text.ConcatString ( Mantissa, Exponent, Buffer );
      ELSIF ProcessedExponent THEN
         IF Ch = '-' THEN
            IF Exponent[1] = '+' THEN
               Exponent[1] := '-';
            ELSE
               Exponent[1] := '+';
            END;
         ELSE
            Exponent[2] := Exponent[3];
            Exponent[3] := Exponent[4];
            Exponent[4] := Ch;
         END;
         Success := Text.ConcatString ( Mantissa, Exponent, Buffer );
      ELSE
         IF Ch = '-' THEN
            IF Mantissa[0] = ' ' THEN
               Mantissa[0] := '-';
            ELSE
               Mantissa[0] := ' ';
            END;
         ELSE
            Success := Text.ConcatChar ( Mantissa, Ch, Mantissa );
            IF Ch = '.' THEN
               DisableKey ( Resource.D9POINT );
            END;
         END;
         Text.Assign ( Mantissa, Buffer );
      END;
      IF NOT (ProcessedExponent OR (Ch = '-')) THEN
         INC ( DigitCount );
      END;
      Success := Text.ConcatChar ( Buffer, ' ', Buffer );
      Readout.Write ( Buffer, UpdateRegion );
      IF DigitCount = MaxDigitsForBase THEN
         DisableKeysABCDEF;
         DisableKeys89;
         DisableKeys234567;
         DisableKey ( Resource.D9KEY1 );
         DisableKey ( Resource.D9KEY0 );
         DisableKey ( Resource.D9POINT );
         KeysDisabled := TRUE;
      END;
   END ConcatChar;

   (********************************************************************)

   PROCEDURE ConvertKeyToCh ( Key    : INTEGER;
                              VAR Ch : CHAR ) : BOOLEAN;

   VAR Success : BOOLEAN;

   BEGIN
      Success := TRUE;
      IF Key = Resource.D9KEYF THEN
         Ch := 'F';
      ELSIF Key = Resource.D9KEYE THEN
         Ch := 'E';
      ELSIF Key = Resource.D9KEYD THEN
         Ch := 'D';
      ELSIF Key = Resource.D9KEYC THEN
         Ch := 'C';
      ELSIF Key = Resource.D9KEYB THEN
         Ch := 'B';
      ELSIF Key = Resource.D9KEYA THEN
         Ch := 'A';
      ELSIF Key = Resource.D9KEY9 THEN
         Ch := '9';
      ELSIF Key = Resource.D9KEY8 THEN
         Ch := '8';
      ELSIF Key = Resource.D9KEY7 THEN
         Ch := '7';
      ELSIF Key = Resource.D9KEY6 THEN
         Ch := '6';
      ELSIF Key = Resource.D9KEY5 THEN
         Ch := '5';
      ELSIF Key = Resource.D9KEY4 THEN
         Ch := '4';
      ELSIF Key = Resource.D9KEY3 THEN
         Ch := '3';
      ELSIF Key = Resource.D9KEY2 THEN
         Ch := '2';
      ELSIF Key = Resource.D9KEY1 THEN
         Ch := '1';
      ELSIF Key = Resource.D9KEY0 THEN
         Ch := '0';
      ELSIF Key = Resource.D9POINT THEN
         Ch := '.';
      ELSE
         Success := FALSE;
      END;
      RETURN (Success);
   END ConvertKeyToCh;

   (********************************************************************)

   PROCEDURE Initialize ();

   VAR Object : Icon.ObjectPtr;

   BEGIN
      Object        := Icon.GetAddress ( Resource.D9CALC, Resource.D9DEC );
      Object^.State := Object^.State + {Icon.Selected};

      Object        := Icon.GetAddress ( Resource.D9CALC, Resource.D9HEX );
      Object^.State := Object^.State - {Icon.Selected};

      Object        := Icon.GetAddress ( Resource.D9CALC, Resource.D9OCT );
      Object^.State := Object^.State - {Icon.Selected};

      Object        := Icon.GetAddress ( Resource.D9CALC, Resource.D9BIN );
      Object^.State := Object^.State - {Icon.Selected};

      Object        := Icon.GetAddress ( Resource.D9CALC, Resource.D9PGM );
      Object^.State := Object^.State - {Icon.Selected};

      NumberBase        := Resource.D9DEC;
      MaxDigitsForBase  := 11;
      InProgramMode     := FALSE;
      KeysDisabled      := FALSE;
      PressedCE         := FALSE;
      DigitsExpected    := 0;
      DigitBuffer       := CHR (0);
      DigitCount        := 0;
      ProcessedExponent := FALSE;
      Mantissa          := " ";
      Exponent          := "E+000";
   END Initialize;

   (********************************************************************)

   PROCEDURE ProcessKey (
      Key          : (* IN  *) INTEGER;
      WindowRegion : (* IN  *) Screen.Box;
      ShowKeys     : (* IN  *) BOOLEAN;
      VAR Number   : (* OUT *) ARRAY OF CHAR ) : BOOLEAN;

   VAR
      ValidToken : BOOLEAN;
      Ch         : CHAR;
      Success    : BOOLEAN;

   BEGIN
      UpdateRegion := WindowRegion;
      DisplayKeys  := ShowKeys;
      ValidToken   := FALSE;
      IF DigitsExpected > 0 THEN

         (*-- The "special" operations (STO, RCL, SUM, GOTO, GTRZER,  --*)
         (*-- and EQLZER) interrupt normal processing to obtain the   --*)
         (*-- required parameter.  Once the required number of digits --*)
         (*-- have been entered, the number is returned and normal    --*)
         (*-- processing resumes.                                     --*)

         IF ConvertKeyToCh ( Key, Ch ) AND (Ch >= '0') AND (Ch <= '9') THEN
            SelectKey   ( Key );   (* "Flash" the key *)
            DeselectKey ( Key );
            Success := Text.ConcatChar ( DigitBuffer, Ch, DigitBuffer );
            DEC ( DigitsExpected );
            IF DigitsExpected = 0 THEN
               ValidToken   := TRUE;
               Text.Assign ( DigitBuffer, Number );
               DeselectKey ( SpecialKey );
               DigitBuffer  := CHR (0);
               EnableDigitKeys ( NumberBase );
               DigitCount   := 0;
            END;
         ELSE
            GEMDOS.ConOut ( Bell );
         END;

      ELSE
         SelectKey   ( Key );   (* "Flash" the key *)
         DeselectKey ( Key );
         IF Key <> Resource.D9CE THEN
            PressedCE := FALSE;
         END;
         IF ConvertKeyToCh ( Key, Ch ) THEN
            ConcatChar ( Ch );
         ELSIF (Key = Resource.D9PLSMIN) AND
               (NumberBase = Resource.D9DEC) AND
               (DigitCount <> 0) THEN
            ConcatChar ( '-' );
         ELSE
            ValidToken := TRUE;
            IF DigitCount = 0 THEN
               Text.Assign ( "", Number );
            ELSE
               IF NumberBase = Resource.D9DEC THEN
                  Success := Text.ConcatString ( Mantissa, Exponent, Number );
               ELSE
                  Text.Assign ( Mantissa, Number );
               END;
            END;
            DigitCount := 0;

            (*--- Do some housekeeping before processing the key ------*)

            IF KeysDisabled THEN
               EnableDigitKeys ( NumberBase );
               EnableKey ( Resource.D9KEY1 );
               EnableKey ( Resource.D9KEY0 );
               KeysDisabled := FALSE;
            END;
            IF NumberBase = Resource.D9DEC THEN
               EnableKey ( Resource.D9POINT );
               DisableKey ( Resource.D9KEYE );
            END;
            IF (Key = Resource.D9DEC) OR
               (Key = Resource.D9HEX) OR
               (Key = Resource.D9OCT) OR
               (Key = Resource.D9BIN) THEN
               SetNumberBase ( Key );
            ELSIF Key = Resource.D9PGM THEN
               ToggleProgramMode;
            ELSIF (Key = Resource.D9STO) OR
                  (Key = Resource.D9RCL) OR
                  (Key = Resource.D9SUM) THEN
               SpecialKey := Key;
               SelectKey ( SpecialKey );
               DigitsExpected := 1;
               EnableDigitKeys ( Resource.D9DEC );
            ELSIF (Key = Resource.D9GOTO) OR
                  (Key = Resource.D9GTRZER) OR
                  (Key = Resource.D9EQLZER) THEN
               SpecialKey := Key;
               SelectKey ( SpecialKey );
               DigitsExpected := 3;
               EnableDigitKeys ( Resource.D9DEC );
            ELSIF Key = Resource.D9CE THEN
               IF NOT PressedCE THEN
                  ClearDisplay ();
                  Readout.Write ( "0 ", UpdateRegion );
                  Number[0]  := CHR (0);
                  ValidToken := FALSE;   (* Dont process the first CE *)
               END;
               PressedCE := TRUE;
            END;
         END;
      END; 
      RETURN (ValidToken);
   END ProcessKey;


END Keypad.

