IMPLEMENTATION MODULE NotePad;

   (*$S-,$T- turn off stack and range checking *) 

   FROM SYSTEM IMPORT ADR, ADDRESS;
   IMPORT M2Conversions;
   IMPORT GEMDOS;
   IMPORT GEMAESbase;
   IMPORT AESResources;
   IMPORT AESEvents;
   IMPORT AESObjects;
   IMPORT AESForms;
   IMPORT AESGraphics;
   IMPORT Text;
   IMPORT Resource;
   IMPORT Screen;
   IMPORT Dialog;
   IMPORT Icon;
   IMPORT Window;
   IMPORT Configuration;
   IMPORT Disk;
   IMPORT File;
   IMPORT Print;
   IMPORT Clock;


   CONST
      MaxLineLength    = 40;
      DoubleLineLength = 80;
      MaxRows          = 9;

      Init    = 1;
      AddChar = 2;
      End     = 3;
      Bell    = 07C;

   TYPE DTA = RECORD
      Reserved   : ARRAY [0..20] OF CHAR;
      Attributes : CHAR;
      Time       : CARDINAL;
      Date       : CARDINAL;
      Size       : LONGCARD;
      Name       : ARRAY [0..13] OF CHAR;
   END;

   TYPE LineStorage = ARRAY [0..MaxLineLength] OF CHAR;
   TYPE LinePtr     = POINTER TO LineStorage;

   TYPE LineType  = RECORD
      ObjectIndex : INTEGER;
      TextPtr     : LinePtr;
   END;

   VAR
      WindowAllocated : BOOLEAN;
      WindowInfo      : Window.Information;
      Line            : ARRAY [0..MaxRows] OF LineType;
      Row             : CARDINAL;
      Column          : INTEGER;
      TreeAddress     : POINTER TO ARRAY [0..1000] OF Icon.Object;
      NumberOfPages   : CARDINAL;
      CurrentPage     : CARDINAL;
      CurrentFileName : Text.String80;
      FileFound       : BOOLEAN;
      InfoLine        : Text.String80;
      NoteChanged     : BOOLEAN;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE NotePadDraw ( Region : Screen.Box );

   BEGIN
      AESObjects.ObjectEdit (
         TreeAddress, Line[Row].ObjectIndex, 0, Column, Init, Column );
   END NotePadDraw;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE SetRow ( NewRow : CARDINAL );

   BEGIN
      AESObjects.ObjectEdit (
         TreeAddress, Line[Row].ObjectIndex, 0, Column, End, Column );
      AESObjects.ObjectEdit (
         TreeAddress, Line[NewRow].ObjectIndex, 0, 0, Init, Column );
      Row := NewRow;
   END SetRow;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE Clear;

   BEGIN
      FOR Row := 0 TO 9 DO
         Line[Row].TextPtr^[0] := CHR (0);
      END (* DO *);
      Row    := 0;
      Column := 0;
   END Clear;

   (************************** LOCAL ROUTINE ***************************)

      PROCEDURE LoadText ( PageNumber : CARDINAL );

   (* This routine loads in the notefile indicated by PageNumber.      *)
   (********************************************************************)

   CONST
      CR         = 15C;
      LF         = 12C;
      Read       = 0;
      NormalFile = 0;
      UseFormat  = TRUE;

   TYPE DateTimeType = RECORD
      Time : CARDINAL;
      Date : CARDINAL;
   END;

   VAR
      SavedDTA     : ADDRESS;
      DTABuffer    : DTA;
      AllNoteFiles : Text.String80;
      Success      : BOOLEAN;
      Result       : INTEGER;
      Index        : CARDINAL;
      FileId       : INTEGER;
      Count        : LONGCARD;
      RowIndex     : CARDINAL;
      ColIndex     : CARDINAL;
      Ch           : CHAR;
      IgnoreChar   : BOOLEAN;
      DateTime     : DateTimeType;
      Buffer       : Text.String80;

   BEGIN
      IF NoteChanged THEN
         SaveText;
      END;
      Clear;

      (*--- Find the filename associated with PageNumber --------------*)

      Index := PageNumber;
      GEMDOS.GetDTA ( SavedDTA );
      GEMDOS.SetDTA ( ADR (DTABuffer) );
      Text.Assign ( Configuration.NoteFileName, AllNoteFiles );
      Success   := Text.ConcatString ( AllNoteFiles, "\NOTE.*", AllNoteFiles );
      Success   := Text.ConcatChar ( Configuration.NoteFileName, '\', CurrentFileName );
      FileFound := FALSE;
      NumberOfPages := 0;
      GEMDOS.SFirst ( AllNoteFiles, 010H, Result );
      LOOP
         IF Result <> 0 THEN
            IF NumberOfPages = 0 THEN
               NumberOfPages := 1;
            END;
            EXIT;
         ELSIF Index > 0 THEN
            DEC (Index);
            IF Index = 0 THEN
               FileFound := TRUE;
               Success   := Text.ConcatString ( CurrentFileName, DTABuffer.Name, CurrentFileName );
            END;
         END;
         INC (NumberOfPages);      
         GEMDOS.SNext ( Result );
      END (* LOOP *);
      GEMDOS.SetDTA ( SavedDTA );
      IF NOT FileFound THEN

         (*--- Make up a file name using PageNumber as extension ------*)

         M2Conversions.ConvertCardinal ( PageNumber, 3, Buffer );
         IF Buffer[0] = ' ' THEN
            Buffer[0] := '0';
         END;
         IF Buffer[1] = ' ' THEN
            Buffer[1] := '0';
         END;
         Success := Text.ConcatString (
            CurrentFileName, "NOTE.", CurrentFileName );
         Success := Text.ConcatString (
            CurrentFileName, Buffer, CurrentFileName );
      END;

      (*--- Read the contents of the file -----------------------------*)

      Disk.CheckStatus ( ORD (CurrentFileName[0]) - ORD ('A') + 1 );
      IF FileFound THEN
         GEMDOS.Open ( CurrentFileName, Read, FileId ); 
      END;
      RowIndex := 0;
      LOOP
         IF (NOT FileFound) OR (FileId < GEMDOS.EOK) THEN
            Line[RowIndex].TextPtr^ := CHR (0);
         ELSE
            IgnoreChar := FALSE;
            ColIndex   := 0;
            LOOP  (* Read the contents of the next line *)
               Count := SIZE (Ch);
               GEMDOS.Read ( FileId, Count, ADR (Ch) );
               IF Count < 1 THEN
                  EXIT;  (* Either error or end-of-file *)
               END;
               IF Ch = CR THEN
                  (* Do nothing *);
               ELSIF Ch = LF THEN
                  Line[RowIndex].TextPtr^[ColIndex] := CHR (0);
                  IgnoreChar := FALSE;
                  EXIT;  (* Go read the next line *)
               ELSIF NOT IgnoreChar THEN
                  Line[RowIndex].TextPtr^[ColIndex] := Ch;
                  INC (ColIndex);
               END;
               IF ColIndex > 39 THEN
                  IgnoreChar := TRUE;
               END;
            END (* LOOP *);
         END;
         INC (RowIndex);
         IF RowIndex > 9 THEN
            EXIT;
         END;
      END (* LOOP *);
      InfoLine := " Note ";
      M2Conversions.ConvertCardinal ( PageNumber, 1, Buffer );
      Success  := Text.ConcatString ( InfoLine, Buffer, InfoLine );
      IF FileFound AND (FileId >= GEMDOS.EOK) THEN
         Success := Text.ConcatString ( InfoLine, " entered on ", InfoLine );
         GEMDOS.DaTime ( FileId, ADR (DateTime), GEMDOS.get );
         Clock.DecodeDate ( DateTime.Date, UseFormat, Buffer );
         Success := Text.ConcatString ( InfoLine, Buffer, InfoLine );
         Success := Text.ConcatString ( InfoLine, " at ", InfoLine );
         Clock.DecodeTime ( DateTime.Time, UseFormat, Buffer );
         Success := Text.ConcatString ( InfoLine, Buffer, InfoLine );
         Success := GEMDOS.Close ( FileId );
      END;
      NoteChanged := FALSE;
      Window.ChangeInformationLine ( WindowInfo.Id, InfoLine );
      IF PageNumber > NumberOfPages THEN
         INC (NumberOfPages);
      END;
      WindowInfo.VirtualRegionSize.Height :=
         WindowInfo.WorkRegion.Size.Height * INTEGER (NumberOfPages);
      Window.SetVirtualRegionSize ( WindowInfo.Id, WindowInfo.VirtualRegionSize );
      IF NumberOfPages > 1 THEN
         Window.MoveVertSlider (
            WindowInfo.Id,
            INTEGER (
               LONGINT (PageNumber - 1) *
               LONGINT (1000) DIV
               LONGINT (NumberOfPages - 1)) );
      END;
   END LoadText;

   (************************** LOCAL ROUTINE ***************************)

      PROCEDURE SaveText;

   (* Save the current page of the notepad in the file NOTE.xxx, where *)
   (* xxx is the number of the CurrentPage.  Note that this routine    *)
   (* may only be called after LoadText.                               *)
   (********************************************************************)

   CONST
      CR         = 15C;
      LF         = 12C;
      NormalFile = 0;

   VAR
      CRLF    : ARRAY [0..1] OF CHAR;
      FileId  : INTEGER;
      Success : BOOLEAN;
      Count   : LONGCARD;
      Index   : CARDINAL;

   BEGIN
      CRLF[0]  := CR;
      CRLF[1]  := LF;
      Disk.CheckStatus ( ORD (CurrentFileName[0]) - ORD ('A') + 1 );
      GEMDOS.Create ( CurrentFileName, NormalFile, FileId ); 
      IF FileId < GEMDOS.EOK THEN
         AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
         File.DosError ( FileId );
      ELSE
         Index := 0;
         LOOP
            Count := LONGCARD (Text.Length ( Line[Index].TextPtr^ ));
            GEMDOS.Write ( FileId, Count, ADDRESS (Line[Index].TextPtr) );
            Count := SIZE (CRLF);
            GEMDOS.Write ( FileId, Count, ADR (CRLF) );
            INC (Index);
            IF Index > 9 THEN
               EXIT;
            END;
         END (* LOOP *);
      END;
      Success     := GEMDOS.Close ( FileId );
      NoteChanged := FALSE;
   END SaveText;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE PrintNote;

   VAR
      Page    : CARDINAL;
      Row     : CARDINAL;

   BEGIN
      IF GEMDOS.PrnOS () THEN
         Page := 1;
         WHILE Page <= NumberOfPages DO
            LoadText ( Page );
            Print.String ( "------------------------------------------" );
            Print.String ( InfoLine );
            Print.String ( "  " );
            FOR Row := 0 TO 9 DO
               Print.String ( Line[Row].TextPtr^ );
            END (* DO *);
            INC (Page);
         END (* DO *);
      END;
      LoadText ( CurrentPage );
   END PrintNote;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE Wrap (
      VAR LongLine : (* IN *) ARRAY OF CHAR;
      RowIndex     : (* IN *) CARDINAL ) : BOOLEAN;

   VAR
      TextPtr    : LinePtr;
      Success    : BOOLEAN;
      HalfLine   : ARRAY [0..DoubleLineLength] OF CHAR;
      DoubleLine : ARRAY [0..DoubleLineLength] OF CHAR;
      Index      : CARDINAL;
 
   BEGIN
      TextPtr := Line[RowIndex].TextPtr;
      IF Text.Length ( LongLine ) <= MaxLineLength THEN

         (*--- The line fits, so stuff it into the notepad ------------*)

         Text.Assign ( LongLine, TextPtr^ );

      ELSIF RowIndex >= MaxRows THEN

         (*--- The line is too long, but there is no more lines to ----*)
         (*--- handle the overflow, so return an error.            ----*)

         RETURN ( FALSE );

      ELSE
         
         (*--- The line is too long, so split the line at the last ----*)
         (*--- full word.                                          ----*)

         Index := MaxLineLength-1;
         LOOP
            IF LongLine[Index] = ' ' THEN
               INC (Index);
               EXIT;
            ELSIF Index = 0 THEN
               EXIT;
            ELSE
               DEC (Index);
            END;
         END (* LOOP *);
         IF NOT Text.SliceString (
            LongLine, Index, Text.Length ( LongLine )-1, HalfLine ) THEN
            RETURN (FALSE);
         END;

         (*--- Wrap the second half of the line to the next row -------*)

         IF NOT Text.ConcatString (
            HalfLine,
            Line[RowIndex+1].TextPtr^,
            DoubleLine ) THEN
            RETURN (FALSE);
         END;
         IF Wrap ( DoubleLine, RowIndex+1 ) THEN

            (*--- Wrap worked, so store the first half of the line ----*)

            IF Index = 0 THEN
               TextPtr^ := CHR (0);
            ELSE
               Success := Text.SliceString ( LongLine, 0, Index-1, HalfLine );
               Text.Assign ( HalfLine, TextPtr^ );
            END;
         ELSE
            RETURN (FALSE);
         END; 
      END;
      RETURN (TRUE);
   END Wrap;

   (********************************************************************)

   PROCEDURE Initialize ();

   TYPE TedInfoPtrType = POINTER TO Icon.Tedinfo;

   VAR
      ObjectPtr  : Icon.ObjectPtr;
      TedinfoPtr : TedInfoPtrType;

   BEGIN
      AESResources.ResourceGetAddr (
         GEMAESbase.RTree, Resource.D15NOTE, TreeAddress );
      Line[0].ObjectIndex := Resource.D15LINE0;
      Line[1].ObjectIndex := Resource.D15LINE1;
      Line[2].ObjectIndex := Resource.D15LINE2;
      Line[3].ObjectIndex := Resource.D15LINE3;
      Line[4].ObjectIndex := Resource.D15LINE4;
      Line[5].ObjectIndex := Resource.D15LINE5;
      Line[6].ObjectIndex := Resource.D15LINE6;
      Line[7].ObjectIndex := Resource.D15LINE7;
      Line[8].ObjectIndex := Resource.D15LINE8;
      Line[9].ObjectIndex := Resource.D15LINE9;
      FOR Row := 0 TO 9 DO
         ObjectPtr  := Icon.GetAddress (
            Resource.D15NOTE, Line[Row].ObjectIndex );
         TedinfoPtr := TedInfoPtrType (ObjectPtr^.Spec);
         Line[Row].TextPtr := LinePtr (TedinfoPtr^.Text);
      END (* DO *);
      Clear;
      NumberOfPages := 1;
   END Initialize;

   (********************************************************************)

   PROCEDURE OwnsWindow ( WindowId : INTEGER ) : BOOLEAN;

   BEGIN
      IF WindowAllocated THEN
         RETURN (WindowId = WindowInfo.Id);
      ELSE
         RETURN (FALSE);
      END;
   END OwnsWindow;

   (********************************************************************)

   PROCEDURE Open;

   VAR
      WindowRegion      : Screen.Box;
      VirtualRegionSize : Window.Area;
      Key               : INTEGER;
      Success           : BOOLEAN;
      FileName          : Text.String80;

   BEGIN
      IF WindowAllocated THEN
         Window.Top ( WindowInfo.Id );
      ELSE
         Icon.GetRegion ( Resource.D15NOTE, 0, WindowRegion );
         WindowRegion.Origin.X := 
            Screen.Center.Origin.X - (WindowRegion.Size.Width DIV 2);
         WindowRegion.Origin.Y := 
            Screen.Center.Origin.Y - (WindowRegion.Size.Height DIV 2);
         VirtualRegionSize.Width  :=
            Window.Pixel (WindowRegion.Size.Width);
         VirtualRegionSize.Height :=
            Window.Pixel (WindowRegion.Size.Height * INTEGER (NumberOfPages));
         IF Window.Open (
               Screen.Center,
               Window.InformationPtr (ADR (WindowInfo)),
               " NotePad ",
               "",
               WindowRegion,
               VirtualRegionSize,
               GEMAESbase.Name +
               GEMAESbase.Closer +
               GEMAESbase.Mover +
               GEMAESbase.InfoLine +
               GEMAESbase.UpArrow +
               GEMAESbase.DownArrow +
               GEMAESbase.VertSlide,
               0, 1,             (* White Background *)
               Resource.D15NOTE,
               NotePadDraw ) THEN
            WindowAllocated := TRUE;
            Row    := 0;
            Column := 0;
            CurrentPage := 1;
            LoadText ( CurrentPage );
         ELSE
            Window.Unavailable;
            WindowAllocated := FALSE;
         END;
      END;
   END Open;

   (********************************************************************)

   PROCEDURE Close;

   BEGIN
      IF WindowAllocated THEN
         AESObjects.ObjectEdit (
            TreeAddress, Line[Row].ObjectIndex, 0, Column, End, Column );
         Window.Close ( WindowInfo.Id, Screen.Center );
         WindowAllocated := FALSE;
      END;
   END Close;

   (********************************************************************)

   PROCEDURE ProcessMessageEvent ( VAR MsgBuffer : ARRAY OF INTEGER ); 

   CONST
      PageUp      = 0;
      PageDown    = 1;
      RowUp       = 2;
      RowDown     = 3;

   VAR
      SliderPosition : INTEGER;
      NewPage        : CARDINAL;

   BEGIN
      IF MsgBuffer[0] = GEMAESbase.AccessoryClose THEN
         IF WindowAllocated THEN
            AESObjects.ObjectEdit (
               TreeAddress, Line[Row].ObjectIndex, 0, Column, End, Column );
            Window.Close ( 0, Screen.Center );
            WindowAllocated := FALSE;
         END;

     ELSIF MsgBuffer[0] = GEMAESbase.WindowArrowed THEN
         AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );

         IF (MsgBuffer[4] = RowUp) OR (MsgBuffer[4] = PageUp) THEN
            IF CurrentPage <= 1 THEN
               GEMDOS.ConOut ( Bell );
            ELSE
               LoadText ( CurrentPage - 1 );
               DEC (CurrentPage);
               Window.Redraw ( WindowInfo.Id, WindowInfo.Borders );
            END;

         ELSIF (MsgBuffer[4] = RowDown) OR (MsgBuffer[4] = PageDown) THEN
            IF CurrentPage >= 999 THEN
               GEMDOS.ConOut ( Bell );
            ELSE
               LoadText ( CurrentPage + 1 );
               INC (CurrentPage);
               Window.Redraw ( WindowInfo.Id, WindowInfo.Borders );
            END;
         END;

      ELSIF MsgBuffer[0] = GEMAESbase.WindowVertSlided THEN
         AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );
         NewPage := CARDINAL (
            LONGCARD (MsgBuffer[4]) *
            LONGCARD (NumberOfPages - 1) DIV
            LONGCARD (1000)) + 1;
         LoadText ( NewPage );
         CurrentPage := NewPage;
         Window.Redraw ( WindowInfo.Id, WindowInfo.Borders );

      ELSE
         GEMDOS.ConOut ( Bell );
      END;
      AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
   END ProcessMessageEvent;

   (********************************************************************)

   PROCEDURE ProcessButtonEvent ( MouseButton : INTEGER;
                                  ModifierKey : INTEGER;
                                  Clicks      : INTEGER );

   VAR
      MouseLocation : Screen.PixelCoordinate;
      MouseState    : INTEGER;
      KeyboardState : INTEGER;
      Object        : INTEGER;
      ObjectPtr     : Icon.ObjectPtr;
      Success       : BOOLEAN;

   BEGIN
      AESGraphics.GrafMouseKeyboardState (
         MouseLocation.X, MouseLocation.Y, MouseState, KeyboardState );
      IF Icon.Find ( Resource.D15NOTE, MouseLocation, Object ) THEN
         ObjectPtr := Icon.GetAddress ( Resource.D15NOTE, Object );
         IF (ObjectPtr^.Type = GEMAESbase.GraphicButton) THEN
            ObjectPtr^.State := ObjectPtr^.State + {Icon.Selected};
            Icon.Display ( Resource.D15NOTE, Object, WindowInfo.Borders );
            ObjectPtr^.State := ObjectPtr^.State - {Icon.Selected};
            AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );
            IF Object = Resource.D15SAVE THEN
               SaveText;
            ELSIF Object = Resource.D15DEL THEN
               Clear;
               Success := GEMDOS.Delete ( CurrentFileName );
               LoadText ( CurrentPage );
               Window.Redraw ( WindowInfo.Id, WindowInfo.Borders );
            ELSIF Object = Resource.D15PRINT THEN
               PrintNote;
            END;
            AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
            Icon.Display ( Resource.D15NOTE, Object, WindowInfo.Borders );
         ELSE
            GEMDOS.ConOut ( Bell );
         END;
      ELSE
         GEMDOS.ConOut ( Bell );
      END;
   END ProcessButtonEvent;

   (********************************************************************)

   PROCEDURE ProcessKeyboardEvent ( Key         : INTEGER;
                                    ModifierKey : INTEGER );

   CONST
      UpArrow    = 48H;
      RightArrow = 4DH;
      DownArrow  = 50H;
      LeftArrow  = 4BH;
      Ins        = 52H;
      Clr        = 47H;
      Del        = 53H;

      BS         = 10C;
      CR         = 15C;

      Col39      = 39;

   VAR
      HighByte      : INTEGER;
      Ch            : CHAR;
      Success       : BOOLEAN;
      RightHalfLine : ARRAY [0..MaxLineLength] OF CHAR;
      LeftHalfLine  : ARRAY [0..MaxLineLength] OF CHAR;
      CurrentLine   : ARRAY [0..DoubleLineLength] OF CHAR;

   BEGIN
      HighByte    := Key DIV 256;
      Ch          := CHAR (Key MOD 256);
      IF (Column = 0) AND (Ch = BS) THEN
         HighByte := UpArrow;
      END;

      IF HighByte = UpArrow THEN
         IF Row = 0 THEN
            GEMDOS.ConOut ( Bell );
         ELSE
            SetRow ( Row - 1 );
         END;

      ELSIF HighByte = DownArrow THEN
         IF Row = 9 THEN
            GEMDOS.ConOut ( Bell );
         ELSE
            SetRow ( Row + 1 );
         END;

      ELSIF HighByte = Clr THEN
         NoteChanged := TRUE;
         AESObjects.ObjectEdit (
            TreeAddress, Line[Row].ObjectIndex, 0, Column, End, Column );
         Clear;
         AESObjects.ObjectEdit (
            TreeAddress, Line[0].ObjectIndex, 0, 0, Init, Column );
         Window.Redraw ( WindowInfo.Id, WindowInfo.Borders );

      ELSIF Ch = CR THEN
         IF Row = 9 THEN
            GEMDOS.ConOut ( Bell );
         ELSE
            SetRow ( Row + 1 );
         END;

      ELSIF (HighByte = LeftArrow) OR
            (HighByte = RightArrow) THEN
         AESObjects.ObjectEdit (
            TreeAddress, Line[Row].ObjectIndex,
            Key,
            Column, AddChar, Column );

      ELSIF (Ch = BS) OR
            (HighByte = Del) OR
            (HighByte = Ins) THEN
         NoteChanged := TRUE;
         AESObjects.ObjectEdit (
            TreeAddress, Line[Row].ObjectIndex,
            Key,
            Column, AddChar, Column );

      ELSE
         NoteChanged := TRUE;
         Text.Assign ( Line[Row].TextPtr^, CurrentLine );
         IF Text.Length ( CurrentLine ) >= MaxLineLength THEN
            IF Column = 0 THEN
               LeftHalfLine[0] := Ch;
               LeftHalfLine[1] := CHR (0);
               Text.Assign ( CurrentLine, RightHalfLine );
            ELSE
               Success := Text.SliceString (
                  CurrentLine, 0, Column-1, LeftHalfLine );
               Success := Text.ConcatChar (
                  LeftHalfLine, Ch, LeftHalfLine );
               Success := Text.SliceString (
                  CurrentLine, Column, Text.Length ( CurrentLine )-1, RightHalfLine );
            END;               
            Success := Text.ConcatString (
               LeftHalfLine, RightHalfLine, CurrentLine );
            IF Wrap ( CurrentLine, Row ) THEN
               IF Column >= MaxLineLength THEN
                  SetRow ( Row + 1 );
               END;
               Window.Redraw ( WindowInfo.Id, WindowInfo.Borders );
            ELSE

               (*--- The notepad is full, so ring the bell ------------*)

               GEMDOS.ConOut ( Bell );
            END;
         ELSE
            AESObjects.ObjectEdit (
               TreeAddress, Line[Row].ObjectIndex,
               Key,
               Column, AddChar, Column );
         END;
      END;
   END ProcessKeyboardEvent;

BEGIN
   WindowAllocated := FALSE;
   NoteChanged     := FALSE;
END NotePad.

