IMPLEMENTATION MODULE File;


   (*$S-,$T- turn off stack and range checking *) 

   FROM SYSTEM IMPORT ADR, ADDRESS, TSIZE;
   IMPORT GEMDOS;
   IMPORT GEMAESbase;
   IMPORT AESForms;
   IMPORT AESResources;
   IMPORT AESObjects;
   IMPORT AESGraphics;
   IMPORT Text;
   IMPORT Resource;
   IMPORT Screen;
   IMPORT Icon;
   IMPORT Configuration;
   IMPORT FileName;
   IMPORT Disk;


   CONST

      DeleteErrorLine1 = "[3][ Can't delete the specified | file. Make sure that the |";
      DeleteErrorLine2 = " file name is valid and the | disk is not write protected.][ OK ]";

      FileExistsLine1 = "[1][ The destination file | already exists. Do you |";
      FileExistsLine2 = " want to delete it? ][ Yes | No ]";

      NotEnoughMemory = "[3][ There is not enough memory | for this operation. ][ OK ]";

      Null = 0;

   VAR
      FileHeaderSize : CARDINAL;
      Banner         : Screen.Box;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE ReadRecord (
      VAR FileId : (* IN *)  FileIdType;
      RecordId   : (* IN *)  RecordIdType;
      Data       : (* OUT *) ADDRESS );

   VAR
      Offset   : LONGINT;
      Position : LONGCARD;
      Count    : LONGCARD;

   BEGIN
      IF RecordId = Null THEN
         Status := EndOfFile;
      ELSE

         (*--- Set the file pointer to the start of the record --------*)

         Offset :=
            LONGINT (RecordId - 1) *
            LONGINT (FileId.RecordSize) +
            LONGINT (FileHeaderSize);
         GEMDOS.Seek (
            Offset, FileId.FileHandle, GEMDOS.beginning, Position );

         (*--- Read the user defined record ---------------------------*)

         Count := LONGCARD (FileId.RecordSize);
         GEMDOS.Read ( FileId.FileHandle, Count, Data );
         IF Count = LONGCARD (FileId.RecordSize) THEN
            Status := EOK;
         ELSE
            Status := ReadFailure;  (* Didn't read entire record *)
         END;
      END;
   END ReadRecord;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE WriteRecord (
      VAR FileId : (* IN *) FileIdType;
      RecordId   : (* IN *) RecordIdType;
      Data       : (* IN *) ADDRESS );
 
   VAR
      Offset   : LONGINT;
      Position : LONGCARD;
      Count    : LONGCARD;

   BEGIN
      IF RecordId = Null THEN
         Status := EndOfFile;
      ELSE

         (*--- Set the file pointer to the start of the record --------*)

         Offset :=
            LONGINT (RecordId - 1) *
            LONGINT (FileId.RecordSize) +
            LONGINT (FileHeaderSize);
         GEMDOS.Seek (
            Offset, FileId.FileHandle, GEMDOS.beginning, Position );

         (*--- Update rest of the record with new user data -----------*)

         Count := LONGCARD (FileId.RecordSize);
         GEMDOS.Write ( FileId.FileHandle, Count, Data );
         IF Count = LONGCARD (FileId.RecordSize) THEN
            Status := EOK;
         ELSE
            Status := WriteFailure;  (* Didn't write entire record *)
         END;
      END;
   END WriteRecord;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE PutField (
      Field      : (* IN *) PossibleFields;
      UserRecord : (* IN *) ADDRESS;
      RecordId   : (* IN *) RecordIdType );

   TYPE RecordHeaderTypePtr = POINTER TO RecordHeaderType;

   VAR HeaderPtr : RecordHeaderTypePtr;

   BEGIN
      HeaderPtr := RecordHeaderTypePtr (UserRecord);
      CASE Field OF
         NextKey    : HeaderPtr^.NextKeyId := RecordId;    |
         PrevKey    : HeaderPtr^.PrevKeyId := RecordId;    |
         NextRecord : HeaderPtr^.NextRecordId := RecordId; |
         PrevRecord : HeaderPtr^.PrevRecordId := RecordId;
      END;
   END PutField;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE UpdateRecord (
      VAR FileId : (* IN *) FileIdType;
      RecordId   : (* IN *) RecordIdType;
      Field      : (* IN *) PossibleFields;
      Data       : (* IN *) RecordIdType;
      WorkBuffer : (* IN *) ADDRESS );

   BEGIN
      IF RecordId <> Null THEN
         ReadRecord ( FileId, RecordId, WorkBuffer );
         IF Status = EOK THEN
            PutField ( Field, WorkBuffer, Data );
            WriteRecord ( FileId, RecordId, WorkBuffer );
         END;
      END;
   END UpdateRecord;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE AllocateRecord (
      VAR FileId   : (* IN *)  FileIdType;
      WorkBuffer   : (* IN *)  ADDRESS;
      VAR RecordId : (* OUT *) RecordIdType );

   VAR
      Position : LONGCARD;
      Count    : LONGCARD;

   BEGIN
      Status := EOK;
      IF FileId.FreeListId = Null THEN
         GEMDOS.Seek ( 0, FileId.FileHandle, GEMDOS.end, Position );
         PutField ( NextRecord, WorkBuffer, Null );
         PutField ( PrevRecord, WorkBuffer, Null );
         Count := LONGCARD (FileId.RecordSize);
         GEMDOS.Write ( FileId.FileHandle, Count, WorkBuffer );
         IF Count = LONGCARD (FileId.RecordSize) THEN
            FileId.FreeListId :=
               (FileId.FileSize - LONGCARD (FileHeaderSize)) DIV 
               LONGCARD (FileId.RecordSize) + 1;
            INC ( FileId.FileSize, LONGCARD (FileId.RecordSize));
            AllocateRecord ( FileId, WorkBuffer, RecordId );
         ELSE
            Status := WriteFailure;
         END;
      ELSE
         RecordId := FileId.FreeListId;
         ReadRecord ( FileId, RecordId, WorkBuffer );
         IF Status = EOK THEN
            FileId.FreeListId := GetField ( NextRecord, WorkBuffer );
            INC (FileId.NumberRecords);
         END;
      END;
   END AllocateRecord;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DisposeRecord (
      VAR FileId : (* IN *) FileIdType;
      RecordId   : (* IN *) RecordIdType;
      WorkBuffer : (* IN *) ADDRESS );

   BEGIN
      PutField ( NextRecord, WorkBuffer, FileId.FreeListId );
      PutField ( PrevRecord, WorkBuffer, Null );
      WriteRecord ( FileId, RecordId, WorkBuffer );
      IF Status = EOK THEN
         FileId.FreeListId := RecordId;
         DEC (FileId.NumberRecords);
      END;
   END DisposeRecord;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE DrawBanner ( Message : INTEGER );

   BEGIN
      Icon.GetRegion ( Message, 0, Banner );
      Banner.Origin.X   := 157;  (* Determined by trial-and-error *)
      Banner.Origin.Y   :=
         Screen.CharacterArea.Height + (Screen.CharacterArea.Height DIV 2);
      Banner.Size.Width := 326;  (* Determined by trial-and-error *)
      Icon.SetOrigin ( Message, 0, Banner.Origin );
      Icon.SetSize ( Message, 0, Banner.Size );
      Icon.Display ( Message, 0, Banner );
   END DrawBanner;

   (************************** LOCAL ROUTINE ***************************)

   PROCEDURE EraseBanner;

   BEGIN
      AESForms.FormDialogue (
         GEMAESbase.FormFinish,
         0, 0, 0, 0,
         Banner.Origin.X,
         Banner.Origin.Y,
         Banner.Size.Width,
         Banner.Size.Height );
   END EraseBanner;

   (********************************************************************)

   PROCEDURE DosError ( ErrorCode : INTEGER );

   VAR
      Button  : INTEGER;
      NewCode : INTEGER;

   BEGIN
      AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
      NewCode := ErrorCode;
      IF NewCode > -50 THEN
         NewCode := NewCode + 31;
         NewCode := -NewCode;
      END;
      Button := AESForms.FormError ( NewCode );
   END DosError;

   (********************************************************************)

   PROCEDURE GetField (
      Field      : (* IN *) PossibleFields;
      UserRecord : (* IN *) ADDRESS ) : RecordIdType;

   TYPE RecordHeaderTypePtr = POINTER TO RecordHeaderType;

   VAR
      HeaderPtr : RecordHeaderTypePtr;
      RecordId  : RecordIdType;

   BEGIN
      HeaderPtr := RecordHeaderTypePtr (UserRecord);
      CASE Field OF
         NextKey    : RecordId := HeaderPtr^.NextKeyId;    |
         PrevKey    : RecordId := HeaderPtr^.PrevKeyId;    |
         NextRecord : RecordId := HeaderPtr^.NextRecordId; |
         PrevRecord : RecordId := HeaderPtr^.PrevRecordId;
      END;
      RETURN (RecordId);
   END GetField;

   (********************************************************************)

   PROCEDURE Create (
      VAR FileName : (* IN *)  ARRAY OF CHAR;
      RecordSize   : (* IN *)  CARDINAL;
      GetKey       : (* IN *)  GetKeyProc;
      GetSubKey    : (* IN *)  GetSubKeyProc;
      VAR FileId   : (* OUT *) FileIdType );

   CONST NormalFile = 0;

   VAR Count : LONGCARD;

   BEGIN

      (*--- Create the file -------------------------------------------*)

      GEMDOS.Create ( FileName, NormalFile, FileId.FileHandle );
      IF FileId.FileHandle < EOK THEN
         Status := FileId.FileHandle;
      ELSE

         (*--- Save the file attributes in the file header ------------*)

         FileId.Valid         := TRUE;
         FileId.FileSize      := LONGCARD (FileHeaderSize);
         FileId.FilePosition  := Null;
         FileId.RecordSize    := RecordSize;
         FileId.NumberRecords := 0;
         FileId.GetKey        := GetKey;
         FileId.GetSubKey     := GetSubKey;
         FileId.FreeListId    := Null;
         FileId.FirstRecordId := Null;
         FileId.NextRecordId  := Null;
         FileId.PrevRecordId  := Null;
         Count := LONGCARD (FileHeaderSize);
         GEMDOS.Write ( FileId.FileHandle, Count, ADR (FileId) );
         IF Count = LONGCARD (FileHeaderSize) THEN
            Status := EOK;
         ELSE
            Status := WriteFailure;
         END;
      END; 
   END Create;

   (********************************************************************)

   PROCEDURE Open (
      VAR FileName   : (* IN *)  ARRAY OF CHAR;
      GetKey         : (* IN *)  GetKeyProc;
      GetSubKey      : (* IN *)  GetSubKeyProc;
      VAR FileId     : (* OUT *) FileIdType );

   CONST ReadWrite = 2;

   VAR Count : LONGCARD;

   BEGIN

      (*--- Open the file ---------------------------------------------*)

      GEMDOS.Open ( FileName, ReadWrite, FileId.FileHandle );
      IF FileId.FileHandle < EOK THEN
         Status := FileId.FileHandle;
      ELSE

         (*--- Get the current file attributes from the header --------*)

         Count := LONGCARD (FileHeaderSize); 
         GEMDOS.Read ( FileId.FileHandle, Count, ADR (FileId) );
         IF Count = LONGCARD (FileHeaderSize) THEN
            FileId.Valid        := TRUE;
            FileId.GetKey       := GetKey;
            FileId.GetSubKey    := GetSubKey;
            FileId.FilePosition := Null;   (* Unknown file position *)
            FileId.NextRecordId := FileId.FirstRecordId;
            Status := EOK;
         ELSE
            Status := ReadFailure;
         END;
      END;
   END Open;

   (********************************************************************)

   PROCEDURE Read (   
      VAR FileId   : (* IN *)  FileIdType;  
      Data         : (* IN *)  ADDRESS;
      VAR RecordId : (* OUT *) RecordIdType );

   BEGIN
      IF NOT FileId.Valid THEN
         Status := InvalidHandle;  (* File is not open *)
      ELSIF FileId.NextRecordId = Null THEN
         Status := EndOfFile;      (* End of file.  *)
      ELSE
         RecordId := FileId.NextRecordId;
         ReadRecord ( FileId, RecordId, Data );
         IF Status = EOK THEN
            FileId.FilePosition := RecordId;
            FileId.NextRecordId := GetField ( NextRecord, Data );
            FileId.PrevRecordId := GetField ( PrevRecord, Data );
         END;
      END;
   END Read;

   (********************************************************************)

   PROCEDURE Write (
      VAR FileId   : (* IN *)  FileIdType;  
      Data         : (* IN *)  ADDRESS;
      VAR RecordId : (* OUT *) RecordIdType );
 
   BEGIN
      IF NOT FileId.Valid THEN
         Status := InvalidHandle;  (* File is not open *)
      ELSIF FileId.FilePosition = Null THEN

         (* Designers Note : The question under consideration is - once *)
         (* we get to the end of the file, should the file be extended? *)
         (* I have elected not to extend the file, but instead return a *)
         (* end-of-file condition.  My reasoning is that by extending   *)
         (* the file, a strong possibility exists for keys contained    *)
         (* in the extended records not being in ascending ACSII order. *)
         (* Extending the file would also muck up the free list.  By    *)
         (* returning an end-of-file condition, the calling program can *)
         (* then call InsertRecord to resolve all of the problems.      *)

         Status := EndOfFile;
      ELSE
         RecordId := FileId.FilePosition;
         WriteRecord ( FileId, RecordId, Data );
         IF Status = EOK THEN
            FileId.FilePosition := GetField ( NextRecord, Data );
            FileId.NextRecordId := GetField ( NextRecord, Data );
            FileId.PrevRecordId := GetField ( PrevRecord, Data );
         END;
      END;
   END Write;

   (********************************************************************)

   PROCEDURE InsertRecord (
      VAR FileId   : (* IN *)  FileIdType;  
      VAR Key      : (* IN *)  ARRAY OF CHAR;
      VAR SubKey   : (* IN *)  ARRAY OF CHAR;
      Data         : (* IN *)  ADDRESS;
      WorkBuffer   : (* IN *)  ADDRESS;
      VAR RecordId : (* OUT *) RecordIdType );

   VAR
      NewRecordId  : RecordIdType;
      NextKeyId    : RecordIdType;
      PrevKeyId    : RecordIdType;
      NextRecordId : RecordIdType;
      PrevRecordId : RecordIdType;
      RecordSubKey : Text.String80;
      NextKeyLink  : RecordIdType;
      PrevKeyLink  : RecordIdType;
      FoundKey     : BOOLEAN;
      FoundRecord  : BOOLEAN;

   BEGIN
      Status := EOK;
      LOOP

         (* Designers Note : The entire body of this routine has been *)
         (* enclosed in a loop in an attempt to simulate exception    *)
         (* handling ala Ada.  If an error is detected, Status will   *)
         (* be set to an error number and the loop will be EXITed.    *)

         AllocateRecord ( FileId, WorkBuffer, NewRecordId );
         IF Status <> EOK THEN
            EXIT;
         END;
         Find (               (* We will use the Find routine to   *)
            FileId,           (* determine where the new record    *)
            Key,              (* should be inserted into the list. *)
            SubKey,           (* We really dont care about the     *)
            WorkBuffer,       (* status code, as long as it doesnt *)
            FoundKey,         (* indicate a read or write error.   *)
            PrevKeyId,
            FoundRecord,
            PrevRecordId );
         IF (Status = EOK) OR (Status = SeekError) OR (Status = EndOfFile) THEN
            Status := EOK;
         ELSE
            EXIT;
         END;
         IF PrevRecordId = Null THEN
         
            (*--- Link the new record into the head of the record list *)
         
            PutField ( NextRecord, Data, FileId.FirstRecordId );
            PutField ( PrevRecord, Data, Null );
            UpdateRecord (
               FileId,
               FileId.FirstRecordId,
               PrevRecord,
               NewRecordId,
               WorkBuffer );

         ELSE
            
            (*--- Link the record as the successor to PrevRecordId ----*)
            
            ReadRecord ( FileId, PrevRecordId, WorkBuffer );
            IF Status <> EOK THEN
               EXIT;
            END;
            NextRecordId := GetField ( NextRecord, WorkBuffer );
            PutField ( NextRecord, Data, NextRecordId );
            PutField ( PrevRecord, Data, PrevRecordId );
            UpdateRecord (
               FileId, NextRecordId, PrevRecord, NewRecordId, WorkBuffer );
            IF Status <> EOK THEN
               EXIT;
            END;
            UpdateRecord (
               FileId, PrevRecordId, NextRecord, NewRecordId, WorkBuffer );

         END;
         IF Status <> EOK THEN
            EXIT;
         END;
         WriteRecord ( FileId, NewRecordId, Data );
         IF Status <> EOK THEN
            EXIT;
         END;
         
         (*--- See if the new record constitutes a new key ------------*)
          
         IF FoundKey THEN
           
            (*--- The key already exists, so see if the new -----------*)
            (*--- record should be linked into the key list -----------*)
            (*--- as the first component.                   -----------*)
            
            ReadRecord ( FileId, PrevKeyId, WorkBuffer );
            IF Status <> EOK THEN
               EXIT;
            END;
            FileId.GetSubKey ( WorkBuffer, RecordSubKey );
            IF Text.Compare ( SubKey, RecordSubKey ) = Text.LessThan THEN
            
               (*--- Link the new record into the key list ------------*)
               (*--- as the new key.                       ------------*)
              
               NextKeyLink := GetField ( NextKey, WorkBuffer );
               PrevKeyLink := GetField ( PrevKey, WorkBuffer );
               PutField ( NextKey, WorkBuffer, Null );
               PutField ( PrevKey, WorkBuffer, Null );
               WriteRecord ( FileId, PrevKeyId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               UpdateRecord ( 
                  FileId, NextKeyLink, PrevKey, NewRecordId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               UpdateRecord (
                  FileId, PrevKeyLink, NextKey, NewRecordId, WorkBuffer );

            ELSE
             
               (*--- The previous record serves as the     ------------*)
               (*--- key, so set the key fields of the new ------------*)
               (*--- record to null.                       ------------*)

               NextKeyLink := Null;
               PrevKeyLink := Null;

            END;
            IF Status <> EOK THEN
               EXIT;
            END;
            ReadRecord ( FileId, NewRecordId, WorkBuffer );
            IF Status <> EOK THEN
               EXIT;
            END;
            PutField ( NextKey, WorkBuffer, NextKeyLink );
            PutField ( PrevKey, WorkBuffer, PrevKeyLink );
            WriteRecord ( FileId, NewRecordId, WorkBuffer );

         ELSE
         
            (*--- The key does not exist, so link the new -------------*)
            (*--- record into the key list as a new key.  -------------*)
           
            IF PrevKeyId = Null THEN

               (*--- Link the new record into the start of the --------*)
               (*--- key list.                                 --------*)

               PutField ( NextKey, Data, FileId.FirstRecordId );
               PutField ( PrevKey, Data, Null );
               UpdateRecord (
                  FileId,
                  FileId.FirstRecordId,
                  PrevKey,
                  NewRecordId,
                  WorkBuffer );

            ELSE

               (*--- Splice the new record into the key list. ---------*)

               ReadRecord ( FileId, PrevKeyId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               NextKeyId := GetField ( NextKey, WorkBuffer );
               PutField ( NextKey, Data, NextKeyId );
               PutField ( PrevKey, Data, PrevKeyId );
               UpdateRecord (
                  FileId, NextKeyId, PrevKey, NewRecordId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               UpdateRecord (
                  FileId, PrevKeyId, NextKey, NewRecordId, WorkBuffer );
            END;
            IF Status <> EOK THEN
               EXIT;
            END;
            WriteRecord ( FileId, NewRecordId, Data );
         END;
         IF Status <> EOK THEN
            EXIT;
         END;
         IF PrevRecordId = Null THEN
            FileId.FirstRecordId := NewRecordId;
         END; 
         RecordId := NewRecordId;
         EXIT;  (* All done! *)
      END (* Exception Handler LOOP *);
   END InsertRecord;

   (********************************************************************)

   PROCEDURE DeleteRecord (   
      VAR FileId : (* IN *) FileIdType;  
      RecordId   : (* IN *) RecordIdType;
      WorkBuffer : (* IN *) ADDRESS );

   VAR
      NextKeyId    : RecordIdType;
      PrevKeyId    : RecordIdType;
      NextRecordId : RecordIdType;
      PrevRecordId : RecordIdType;
      Key          : Text.String80;
      Successor    : Text.String80;
      MoveKeyInfo  : BOOLEAN;

   BEGIN
      Status := EOK;
      LOOP

         (* Designers Note : The entire body of this routine has been *)
         (* enclosed in a loop in an attempt to simulate exception    *)
         (* handling ala Ada.  If an error is detected, Status will   *)
         (* be set to an error number and the loop will be EXITed.    *)

         ReadRecord ( FileId, RecordId, WorkBuffer );
         IF Status <> EOK THEN
            EXIT;
         END;
         NextKeyId    := GetField ( NextKey, WorkBuffer );
         PrevKeyId    := GetField ( PrevKey, WorkBuffer );
         NextRecordId := GetField ( NextRecord, WorkBuffer );
         PrevRecordId := GetField ( PrevRecord, WorkBuffer );
         FileId.GetKey ( WorkBuffer, Key );
         DisposeRecord ( FileId, RecordId, WorkBuffer );
         IF Status <> EOK THEN
            EXIT;
         END;

         (*--- Cut the record from the record list. -------------------*)

         UpdateRecord (
            FileId, NextRecordId, PrevRecord, PrevRecordId, WorkBuffer );
         IF Status <> EOK THEN
            EXIT;
         END;
         UpdateRecord (
            FileId, PrevRecordId, NextRecord, NextRecordId, WorkBuffer );
         IF Status <> EOK THEN
            EXIT;
         END;
         IF FileId.FirstRecordId = RecordId THEN
            FileId.FirstRecordId := NextRecordId;
         END;

         (*--- See if the key record list should be updated. ----------*)

         IF (PrevKeyId = Null) AND (NextKeyId = Null) THEN

            (* Do nothing, since the record is not a key record *)

         ELSE

            (*--- A key record is being deleted.  See if the ----------*)
            (*--- next record should become the new key record. -------*)

            IF NextRecordId = Null THEN
               MoveKeyInfo := FALSE;
            ELSE
               ReadRecord ( FileId, NextRecordId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               FileId.GetKey ( WorkBuffer, Successor );
               MoveKeyInfo :=
                  (Text.Compare ( Key, Successor ) = Text.Equal);
            END;

            IF MoveKeyInfo THEN
               UpdateRecord (
                  FileId, NextRecordId, PrevKey, PrevKeyId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               UpdateRecord (
                  FileId, NextRecordId, NextKey, NextKeyId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               UpdateRecord (
                  FileId, PrevKeyId, NextKey, NextRecordId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               UpdateRecord (
                  FileId, NextKeyId, PrevKey, NextRecordId, WorkBuffer );

            ELSE

               (*--- The deleted record is not linked to any ----------*)
               (*--- subrecords, so remove the key from the  ----------*)
               (*--- key list.                               ----------*)

               UpdateRecord (
                  FileId, PrevKeyId, NextKey, NextKeyId, WorkBuffer );
               IF Status <> EOK THEN
                  EXIT;
               END;
               UpdateRecord (
                  FileId, NextKeyId, PrevKey, PrevKeyId, WorkBuffer );
            END;
         END;
         EXIT;  (* All done! *)
      END (* Exception Handler LOOP *) ;
   END DeleteRecord;

   (********************************************************************)

   PROCEDURE Retreat (
      VAR FileId   : (* IN *)  FileIdType;
      Data         : (* OUT *) ADDRESS;  
      VAR RecordId : (* OUT *) RecordIdType );

   BEGIN
      IF NOT FileId.Valid THEN
         Status  := InvalidHandle;  (* File is not open *)
      ELSIF FileId.PrevRecordId = Null THEN
         Status  := SeekError;      (* Cant retreat past start of file *)
      ELSE
         Seek ( FileId, FileId.PrevRecordId, Data );
      END;
   END Retreat;

   (********************************************************************)

   PROCEDURE Seek (
      VAR FileId : (* IN *)  FileIdType;  
      RecordId   : (* IN *)  RecordIdType;
      Data       : (* OUT *) ADDRESS );

   BEGIN
      IF NOT FileId.Valid THEN
         Status := InvalidHandle;  (* File is not open *)
      ELSE
         ReadRecord ( FileId, RecordId, Data );
         IF Status = EOK THEN
            FileId.FilePosition := RecordId;
            FileId.NextRecordId := GetField ( NextRecord, Data );
            FileId.PrevRecordId := GetField ( PrevRecord, Data );
         END;
      END;
   END Seek;

   (********************************************************************)

   PROCEDURE Find (
      VAR FileId      : (* IN *)     FileIdType;
      VAR Key         : (* IN *)     ARRAY OF CHAR;
      VAR SubKey      : (* IN *)     ARRAY OF CHAR;  
      Data            : (* IN OUT *) ADDRESS;
      VAR FoundKey    : (* OUT *)    BOOLEAN;
      VAR KeyId       : (* OUT *)    RecordIdType;
      VAR FoundRecord : (* OUT *)    BOOLEAN;
      VAR RecordId    : (* OUT *)    RecordIdType );

   VAR
      PrevKeyId     : RecordIdType;
      PrevRecordId  : RecordIdType;
      RecordKey     : Text.String80;
      RecordSubKey  : Text.String80;
      CompareResult : Text.CompareResult;

   BEGIN
      KeyId       := Null;
      RecordId    := Null;
      FoundKey    := FALSE;
      FoundRecord := FALSE;
      IF NOT FileId.Valid THEN
         Status := InvalidHandle;  (* File is not open *)
      ELSIF FileId.FirstRecordId = Null THEN
         Status := SeekError;      (* No records in file *)
      ELSE
         PrevKeyId := Null;
         KeyId     := FileId.FirstRecordId;

         (*--- Scan down the key list.  The loop terminates once the --*)
         (*--- key has been found, or once the key hasn't been found --*)
         (*--- under the following conditions:                       --*)
         (*---                                                       --*)
         (*---    1) An error was detected while reading a record.   --*)
         (*---    2) The key just read is greater than the key we    --*)
         (*---       are looking for.                                --*)
         (*---    3) The end-of-file was detected before the key     --*)
         (*---       was found.                                      --*)

         LOOP
            ReadRecord ( FileId, KeyId, Data );
            IF Status = EndOfFile THEN
               KeyId := PrevKeyId;

               (*--- Found the last key without a match, so scan  -----*)
               (*--- down the rest of the records until found the -----*)
               (*--- last record.                                 -----*)

               PrevRecordId := KeyId;
               RecordId     := KeyId;
               LOOP
                  ReadRecord ( FileId, RecordId, Data );
                  IF Status = EOK THEN
                     PrevRecordId := RecordId;
                     RecordId     := GetField ( NextRecord, Data );
                  ELSE
                     RecordId := PrevRecordId;
                     EXIT;
                  END;
               END; (* LOOP *)
               EXIT;
            ELSIF Status <> EOK THEN
               EXIT;
            END;

            (*--- Still looking for the key, so compare the keys ------*)

            FileId.GetKey ( Data, RecordKey );
            CompareResult := Text.Compare ( RecordKey, Key );
            IF CompareResult = Text.LessThan THEN

               (*--- Get the next key record for comparision ----------*)

               PrevKeyId := KeyId;
               KeyId     := GetField ( NextKey, Data );

            ELSIF CompareResult = Text.Equal THEN

               (*--- Found the key, but not done yet! -----------------*)

               FoundKey := TRUE;
               RecordId := KeyId;
               EXIT;

            ELSE (* CompareResult = Text.GreaterThan *)

               (*--- Went too far, so backup one key ------------------*)

               KeyId    := PrevKeyId;
               RecordId := GetField ( PrevRecord, Data );
               EXIT;
            END;
         END; (* LOOP *)
         IF (Status = EOK) AND (FoundKey) THEN

            (*--- Found the key, so now look for the subkey.  The -----*)
            (*--- loop terminates once the subkey has been found, -----*)
            (*--- or once the subkey hasn't been found under the  -----*)
            (*--- following conditions:                           -----*)
            (*---                                                 -----*)
            (*---    1) An error was detected while reading a     -----*)
            (*---       record.                                   -----*)
            (*---    2) The subkey just read is greater than the  -----*)
            (*---       subkey we are looking for.                -----*)
            (*---    3) The end-of-file was detected before the   -----*)
            (*---       subkey was found.                         -----*)

            LOOP
               FileId.GetSubKey ( Data, RecordSubKey );
               CompareResult := Text.Compare ( RecordSubKey, SubKey );
               IF CompareResult = Text.LessThan THEN

                  (*--- Get the next record for comparision -----------*)

                  PrevRecordId := RecordId;
                  RecordId     := GetField ( NextRecord, Data );
                  ReadRecord ( FileId, RecordId, Data );
                  IF Status = EndOfFile THEN

                     (*--- We are at the end of file, so back up ------*)
                     (*--- one record and exit the loop.         ------*)

                     RecordId := PrevRecordId;
                     EXIT;
                  ELSIF Status <> EOK THEN
                     EXIT;
                  END;

                  (*--- Make sure we still are using the same ---------*)
                  (*--- primary key.                          ---------*)

                  FileId.GetKey ( Data, RecordKey );
                  IF Text.Compare ( RecordKey, Key ) <> Text.Equal THEN

                     (*--- We are now using a different primary -------*)
                     (*--- key, so backup one record and exit   -------*)
                     (*--- the loop.                            -------*)

                     RecordId := PrevRecordId;
                     EXIT;
                  END;

               ELSIF CompareResult = Text.Equal THEN

                  (*--- Found the record we were looking for! ---------*)

                  FoundRecord := TRUE;
                  EXIT;

               ELSE (* CompareResult = Text.GreaterThan *)

                  (*--- We went too far, so backup one record ---------*)
                  (*--- and exit the loop.                    ---------*)

                  RecordId := GetField ( PrevRecord, Data );
                  EXIT;
               END;
            END; (* LOOP *)
         END;
      END;
   END Find;

   (********************************************************************)

   PROCEDURE Close ( VAR FileId : (* IN *) FileIdType );

   VAR Count : LONGCARD;

   BEGIN
      IF NOT FileId.Valid THEN
         Status := InvalidHandle;   (* File is not open *)
      ELSE
         GEMDOS.Seek (
            0, FileId.FileHandle, GEMDOS.beginning, FileId.FilePosition );
         FileId.Valid := FALSE;
         Count        := LONGCARD (FileHeaderSize);
         GEMDOS.Write ( FileId.FileHandle, Count, ADR (FileId) );
         IF Count = LONGCARD (FileHeaderSize) THEN
            IF GEMDOS.Close ( FileId.FileHandle ) THEN
               Status := EOK;
            ELSE
               Status := UnknownError;
            END;
         ELSE
            Status := WriteFailure;
         END;
      END;
   END Close;

   (********************************************************************)

   PROCEDURE Copy;

   CONST
      Cancel     = 0;
      ReadWrite  = 2;
      NormalFile = 0;

   VAR
      Directory       : Text.String80;
      FileNameBody    : Text.String80;
      SourceFileName  : Text.String80;
      DestFileName    : Text.String80;
      ExitButton      : INTEGER;
      Success         : BOOLEAN;
      Button          : INTEGER;
      ErrorMsg        : ARRAY [0..150] OF CHAR;
      Buffer          : ADDRESS;
      MaxBufferSize   : LONGCARD;
      BufferLength    : LONGCARD;
      Abort           : BOOLEAN;
      MemoryAllocated : BOOLEAN;
      SourceId        : INTEGER;
      DestId          : INTEGER;
      Count           : LONGCARD;
      Choice          : INTEGER;
 
   BEGIN
      Abort           := FALSE;
      MemoryAllocated := FALSE;

      (*--- Allocate the disk buffer memory ---------------------------*)

      GEMDOS.Alloc ( 0FFFFFFFFH, Buffer );
      MaxBufferSize := LONGCARD (Buffer);
      IF MaxBufferSize >= 4000H THEN
         MaxBufferSize := MaxBufferSize - 3000H;  (* leave space for File Selector *)
         GEMDOS.Alloc ( MaxBufferSize, Buffer );
         MemoryAllocated := TRUE;
      ELSE
         Choice := AESForms.FormAlert ( 1, NotEnoughMemory );
         Abort  := TRUE;
      END;
 
      (*--- Get the source filename -----------------------------------*)

      IF NOT Abort THEN
         Text.Assign ( Configuration.DefaultDrive, Directory );
         FileNameBody := "";
         DrawBanner ( Resource.D16SRC );
         AESForms.FileSelectorInput ( ADR (Directory), ADR (FileNameBody), ExitButton );
         EraseBanner;
         IF ExitButton = Cancel THEN
            Abort := TRUE;
         ELSIF Text.Length (FileNameBody) = 0 THEN
            Abort := TRUE;
         ELSE
            FileName.BuildFileName ( Directory, FileNameBody, SourceFileName );
         END;
      END;

      (*--- Get the destination filename ------------------------------*)

      IF NOT Abort THEN
         Text.Assign ( Configuration.DefaultDrive, Directory );
         FileNameBody  := "";
         DrawBanner ( Resource.D17DEST );
         AESForms.FileSelectorInput ( ADR (Directory), ADR (FileNameBody), ExitButton );
         EraseBanner;
         IF ExitButton = Cancel THEN
            Abort := TRUE;
         ELSIF Text.Length (FileNameBody) = 0 THEN
            Abort := TRUE;
         ELSE
            FileName.BuildFileName ( Directory, FileNameBody, DestFileName );
         END;
      END;

      (*--- Check to see if the destination file exists ---------------*)

      IF NOT Abort THEN
         AESGraphics.GrafMouse ( GEMAESbase.HourGlass, NIL );
         GEMDOS.Open ( DestFileName, ReadWrite, DestId );
         IF DestId >= GEMDOS.EOK THEN
            ErrorMsg := FileExistsLine1;
            Success  := Text.ConcatString ( ErrorMsg, FileExistsLine2, ErrorMsg );
            AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
            Abort    := (AESForms.FormAlert ( 2, ErrorMsg ) = 2);
            Success  := GEMDOS.Close ( DestId );
         END;
      END;

      (*--- Open the source and destination files ---------------------*)

      SourceId := GEMDOS.EIHndl;
      DestId   := GEMDOS.EIHndl;
      IF NOT Abort THEN
         GEMDOS.Open ( SourceFileName, ReadWrite, SourceId );
         IF SourceId < GEMDOS.EOK THEN
            DosError ( SourceId );
            Abort := TRUE;
         END;
      END;
      IF NOT Abort THEN
         GEMDOS.Create ( DestFileName, NormalFile, DestId ); 
         IF DestId < GEMDOS.EOK THEN
            DosError ( DestId );
            Abort := TRUE;
         END;
      END;

      (*--- Transfer the bytes! ---------------------------------------*)

      IF NOT Abort THEN
         LOOP
            Count := MaxBufferSize;
            GEMDOS.Read ( SourceId, Count, Buffer );
            IF Count = 0 THEN
               EXIT;
            ELSE
               GEMDOS.Write ( DestId, Count, Buffer );
            END;
         END (* LOOP *);
      END;

      (*--- All done, so close the files ------------------------------*)

      IF SourceId >= GEMDOS.EOK THEN
         Success := GEMDOS.Close ( SourceId );
      END;
      IF DestId >= GEMDOS.EOK THEN
         Success := GEMDOS.Close ( DestId );
      END;

      (*--- Free up the memory ----------------------------------------*)

      IF MemoryAllocated THEN
         Success := GEMDOS.Free ( ADDRESS (Buffer) );
      END;
      AESGraphics.GrafMouse ( GEMAESbase.Arrow, NIL );
   END Copy;

   (********************************************************************)

   PROCEDURE Delete;

   CONST Cancel = 0;

   VAR
      Directory    : Text.String80;
      FileNameBody : Text.String80;
      FullFileName : Text.String80;
      ExitButton   : INTEGER;
      Success      : BOOLEAN;
      Button       : INTEGER;
      ErrorMsg     : ARRAY [0..150] OF CHAR;

   BEGIN
      Text.Assign ( Configuration.DefaultDrive, Directory );
      FileNameBody  := "";
      LOOP
         AESForms.FileSelectorInput ( ADR (Directory), ADR (FileNameBody), ExitButton );
         IF ExitButton = Cancel THEN
            EXIT;
         ELSIF Text.Length (FileNameBody) > 0 THEN
            FileName.BuildFileName ( Directory, FileNameBody, FullFileName );
            IF FullFileName[1] = ':' THEN
               Disk.CheckStatus ( ORD (FullFileName[0]) - ORD ('A') + 1 );
            END;
            IF NOT GEMDOS.Delete ( FullFileName ) THEN
               ErrorMsg := DeleteErrorLine1;
               Success  := Text.ConcatString ( ErrorMsg, DeleteErrorLine2, ErrorMsg );
               Button   := AESForms.FormAlert ( 1, ErrorMsg );
            END;
         END;
      END (* LOOP *);
   END Delete;


BEGIN
   FileHeaderSize := TSIZE (FileIdType);
END File.

