IMPLEMENTATION MODULE Screen;


   (*$S-,$T- turn off stack and range checking *) 

   IMPORT Math;
   IMPORT XBIOS;
   IMPORT GEMVDIbase;
   IMPORT GEMAESbase;
   IMPORT VDIInquires;
   IMPORT VDIControls;
   IMPORT AESGraphics;
   IMPORT AESWindows;


   (********************************************************************)

   PROCEDURE Open;

   CONST
      RasterCoordinateSystem = 2;
      Desktop                = 0;

   VAR
      CharacterSize  : Area;
      Index          : CARDINAL;
      WorkIn         : GEMVDIbase.VDIWorkInType;
      WorkOut        : GEMVDIbase.VDIWorkOutType;
      TextAttributes : GEMVDIbase.TextAttrArrayType;
      ScreenType     : INTEGER;

   BEGIN

      (***** Obtain desktop handle and then open the workstation *****)

      PhysicalHandle := AESGraphics.GrafHandle (
         CharacterSize.Width,
         CharacterSize.Height,
         CharacterArea.Width,
         CharacterArea.Height );
      VirtualHandle := PhysicalHandle;
      FOR Index := 0 TO 9 DO
         WorkIn[Index] := 1
      END;
      WorkIn[10] := RasterCoordinateSystem;
      VDIControls.OpenVirtualWorkstation (
         WorkIn, VirtualHandle, WorkOut );
      Center.Origin.X    := (WorkOut[0] + 1) DIV 2;
      Center.Origin.Y    := (WorkOut[1] + 1) DIV 2;
      Center.Size.Width  := 1;
      Center.Size.Height := 1;
      ScreenType := XBIOS.GetResolution ();
      IF ScreenType = 0 THEN
         Resolution := LowRes;
      ELSIF ScreenType = 1 THEN
         Resolution := MediumRes;
      ELSE
         Resolution := HighRes;
      END;
      VDIInquires.InquireTextAttributes ( VirtualHandle, TextAttributes );
      CharacterArea.Width  := TextAttributes[8];
      CharacterArea.Height := TextAttributes[9];
      Center.Origin.Y := Center.Origin.Y + CharacterArea.Height;
      AESWindows.WindowGet (
         Desktop,
         GEMAESbase.WorkXYWH,
         WorkSpace.Origin.X, 
         WorkSpace.Origin.Y,
         WorkSpace.Size.Width,
         WorkSpace.Size.Height ); 
   END Open;

   (********************************************************************)

   PROCEDURE Close;

   BEGIN
      VDIControls.CloseVirtualWorkstation ( VirtualHandle );
   END Close;

   (********************************************************************)

   PROCEDURE SetClippingRectangle ( ClipFlag  : ClipStatus;
                                    Rectangle : Box );

   VAR
      PxyArray : GEMVDIbase.PxyArrayType;

   BEGIN
      PxyArray[0] := Rectangle.Origin.X;
      PxyArray[1] := Rectangle.Origin.Y;
      PxyArray[2] := Rectangle.Origin.X + Rectangle.Size.Width - 1;
      PxyArray[3] := Rectangle.Origin.Y + Rectangle.Size.Height - 1;
      VDIControls.SetClipping ( VirtualHandle, ORD (ClipFlag), PxyArray );
   END SetClippingRectangle;


   (********************************************************************)

   PROCEDURE ContainsPoint ( Region : Box;
                             Point  : PixelCoordinate ) : BOOLEAN;

   BEGIN
      RETURN ( (Point.X >= Region.Origin.X) AND
               (Point.Y >= Region.Origin.Y) AND
               (Point.X < Region.Origin.X + Region.Size.Width) AND
               (Point.Y < Region.Origin.Y + Region.Size.Height) );
   END ContainsPoint;

   (********************************************************************)

   PROCEDURE ContainsRectangle ( Region    : Box;
                                 Rectangle : Box ) : BOOLEAN;

   BEGIN
      RETURN ( (Rectangle.Origin.X >= Region.Origin.X) AND
               (Rectangle.Origin.Y >= Region.Origin.Y) AND
               ((Rectangle.Origin.X + Rectangle.Size.Width) <=
                  (Region.Origin.X + Region.Size.Width)) AND
               ((Rectangle.Origin.Y + Rectangle.Size.Height) <=
                  (Region.Origin.Y + Region.Size.Height)) );
   END ContainsRectangle;

   (********************************************************************)

   PROCEDURE Intersection ( Region1     : Box;
                            Region2     : Box;
                            VAR Overlap : Box ) : BOOLEAN;

   VAR Temp : Box;

   BEGIN
      Temp.Origin.X := Math.Max ( Region1.Origin.X, Region2.Origin.X );
      Temp.Origin.Y := Math.Max ( Region1.Origin.Y, Region2.Origin.Y );
      Temp.Size.Width :=
         Math.Min ( Region1.Origin.X + Region1.Size.Width,
                    Region2.Origin.X + Region2.Size.Width ) -
         Temp.Origin.X;
      Temp.Size.Height :=
         Math.Min ( Region1.Origin.Y + Region1.Size.Height,
                    Region2.Origin.Y + Region2.Size.Height ) -
         Temp.Origin.Y;
      Overlap := Temp;
      RETURN ( (Overlap.Size.Width > 0) AND (Overlap.Size.Height > 0) );
   END Intersection;


END Screen.

